import json
from abc import ABCMeta
from io import BytesIO
from pathlib import Path
import numpy as np
from biotite.file import File, InvalidFileError
from biotite.structure import AtomArray, AtomArrayStack
from ... import color
from ...assets import data
from ...utils import count_value_changes
from . import selections


class ReaderBase(metaclass=ABCMeta):
    """
    Abstract base class for reading molecular data from a file.
    """

    def __init__(self, file_path: str | Path | BytesIO):
        self._extra_annotations: dict
        self.file_path = file_path
        self.file = self.read(file_path)
        self.array: AtomArray | AtomArrayStack = self.get_structure()
        self.array = self.set_extra_annotations(
            self._extra_annotations, self.array, self.file
        )
        self.array = self.set_standard_annotations(self.array)

    @property
    def n_models(self) -> int:
        return self.array.stack_depth()  # type: ignore

    def read(self, file_path: str | Path | BytesIO) -> File:
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def set_standard_annotations(
        cls, array: AtomArray | AtomArrayStack
    ) -> AtomArray | AtomArrayStack:
        annotations = {
            "mass": cls._compute_mass,
            "atomic_number": cls._compute_atomic_number,
            "res_name_int": cls._compute_res_name_int,
            "chain_id_int": cls._compute_chain_id_int,
            "ures_id": cls._compute_ures_id,
            "vdw_radii": cls._compute_vdw_radii,
            "atom_name_int": cls._compute_atom_name_int,
            "charge": cls._compute_charge,
            "lipophobicity": cls._compute_lipophobicity,
            "Color": cls._compute_color,
            "is_alpha_carbon": cls._compute_is_alpha_carbon,
            "is_solvent": cls._compute_is_solvent,
            "is_backbone": cls._compute_is_backbone,
            "is_nucleic": cls._compute_is_nucleic,
            "is_peptide": cls._compute_is_peptide,
            "is_hetero": cls._compute_is_hetero,
            "is_side_chain": cls._compute_is_side_chain,
            "is_carb": cls._compute_is_carb,
        }
        for key, func in annotations.items():
            try:
                array.set_annotation(key, func(array))
            except Exception:
                pass

        return array

    def get_structure(self, model: int | None = None) -> AtomArrayStack | AtomArray:
        raise NotImplementedError("Subclasses must implement this method.")

    def entity_ids(self) -> list[str] | None:
        return None

    def chain_ids(self) -> list[str] | None:
        return np.unique(self.array.get_annotation("chain_id")).tolist()

    def assemblies(self, as_json_string: bool = False) -> dict | str:
        try:
            if as_json_string:
                return json.dumps(self._assemblies())
            return self._assemblies()
        except InvalidFileError:
            return ""

    def _assemblies(self):
        return {}

    @staticmethod
    def set_extra_annotations(
        annotations: dict, array, file
    ) -> AtomArray | AtomArrayStack:
        for name, func in annotations.items():
            try:
                # for the getting of some custom attributes, we get it for the full atom_site
                # but we need to assign a subset of the array that is that model in the full
                # atom_site, so we subset using the atom_id
                try:
                    array.set_annotation(name, func(array, file))
                except IndexError:
                    array.set_annotation(
                        name,
                        func(array, file)[array.atom_id - 1],  # type: ignore
                    )
            except KeyError:
                pass
                # if True:
                #     print(f"Unable to add {name} as an attribute, error: {e}")

        return array

    @staticmethod
    def _compute_mass(array):
        return np.array(
            [
                data.elements.get(x, {}).get("standard_mass", 0.0)
                for x in np.char.title(array.element)
            ],
            dtype=float,
        )

    @staticmethod
    def _compute_atomic_number(array):
        return np.array(
            [
                data.elements.get(x, {}).get("atomic_number", 0)
                for x in np.char.title(array.element)
            ],
            dtype=int,
        )

    @staticmethod
    def _compute_res_name_int(array):
        return np.array(
            [
                data.residues.get(name, {}).get("res_name_num", -1)
                for name in array.res_name
            ],
            dtype=int,
        )

    @staticmethod
    def _compute_chain_id_int(array):
        if isinstance(array.chain_id[0], int):
            return array.chain_id.astype(int)
        else:
            return np.unique(array.chain_id, return_inverse=True)[1].astype(int)

    @staticmethod
    def _compute_vdw_radii(array):
        return np.array(
            [
                data.elements.get(x, {}).get("vdw_radii", 100.0) / 100
                for x in np.char.title(array.element)
            ],
            dtype=float,
        )

    @staticmethod
    def _compute_atom_name_int(array):
        return np.array(
            [data.atom_names.get(x, -1) for x in array.atom_name], dtype=int
        )

    @staticmethod
    def _compute_charge(array):
        return np.array(
            [
                data.atom_charge.get(res, {}).get(atom, 0)
                for res, atom in zip(array.res_name, array.atom_name)
            ],
            dtype=float,
        )

    @staticmethod
    def _compute_lipophobicity(array):
        return np.array(
            [
                data.lipophobicity.get(res, {}).get(atom, 0)
                for res, atom in zip(array.res_name, array.atom_name)
            ],
            dtype=float,
        )

    @staticmethod
    def _compute_color(array, color_plddt: bool = False):
        if color_plddt:
            return color.plddt(array.b_factor)
        else:
            return color.color_chains(array.atomic_number, array.chain_id)

    @staticmethod
    def _compute_is_alpha_carbon(array):
        return selections.select_alpha_carbon(array)

    @staticmethod
    def _compute_is_hetero(array):
        return array.hetero

    @staticmethod
    def _compute_is_backbone(array):
        return selections.select_backbone(array)

    @staticmethod
    def _compute_is_solvent(array):
        return selections.select_solvent(array)

    @staticmethod
    def _compute_is_nucleic(array):
        return selections.select_nucleotides(array)

    @staticmethod
    def _compute_is_peptide(array):
        return selections.select_peptide(array)

    @staticmethod
    def _compute_is_side_chain(array):
        return selections.select_side_chain(array)

    @staticmethod
    def _compute_is_carb(array):
        return selections.select_carbohydrates(array)

    @staticmethod
    def _compute_ures_id(array):
        return count_value_changes(array.res_id, array.chain_id_int)



import databpy as db
import numpy as np
from MDAnalysis import Universe
from MDAnalysis.coordinates.base import ReaderBase
from MDAnalysis.core.topology import Topology
from MDAnalysis.core.topologyattrs import (
    Atomids,
    Bonds,
    ChainIDs,
    Resids,
    Resnames,
)

# Resnums,
from MDAnalysis.lib import util
from MDAnalysis.topology.base import TopologyReaderBase
from ... import color
from ...blender import coll
from ...nodes import nodes
from ..base import EntityType
from .base import Trajectory

DNA_SCALE = 10


class OXDNAParser(TopologyReaderBase):
    def parse(self, **kwargs):
        """Parse topology from an oxDNA topology file.

        Returns
        -------
        top : MDAnalysis.core.topology.Topology
            Topology object
        """
        top = self._parseatoms()

        return top

    @classmethod
    def _is_new_topology(cls, filename) -> bool:
        """Check if topology file is new format.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        bool
            True if topology file is new format
        """
        with open(filename) as f:
            return "5->3" in f.readline()

    @classmethod
    def _read_topo_new(cls, filename) -> Topology:
        """Read topology from new format oxDNA topology file.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        with open(filename) as f:
            lines = f.readlines()
        n_atoms, n_chains, direction = np.array(lines[0].split())
        n_atoms = int(n_atoms)
        atom_idx = np.arange(n_atoms)

        res_name_list = []
        chain_id_list = []
        for i, line in enumerate(lines[1:]):
            is_rna = "type=RNA" in line
            _is_dna = not is_rna

            line_split = line.split()
            bases = line_split[0]

            base_list = []
            start = 0
            end = 0
            in_custom_base = False
            for i, letter in enumerate(bases):
                if letter == "(":
                    start = i + 1
                    in_custom_base = True
                    continue
                if letter == ")":
                    end = i
                    base_list.append(line[start:end])
                    in_custom_base = False
                    continue

                if in_custom_base:
                    continue

                base_list.append(letter)
            chain_id_list.append(np.repeat(i, len(base_list)))

            res_name_list.append(np.array(base_list))

        res_names = np.hstack(res_name_list)
        chain_ids = np.hstack(chain_id_list)

        bond_idx = np.zeros((n_atoms, 2), dtype=int)
        bond_idx[:, :] = -1

        for i in atom_idx:
            if i == 1:
                continue
            if chain_ids[i] == chain_ids[-1]:
                bond_idx[i, :] = np.array((i, i - 1), dtype=int)

        mask = np.logical_and(bond_idx[:, 0] != -1, bond_idx[:, 1] != -1)
        bond_idx = bond_idx[mask, :]

        attrs = [
            Atomids(atom_idx),
            Resids(atom_idx + 1),
            ChainIDs(chain_ids),
            Resnames(res_names),
        ]

        topo = Topology(
            n_atoms=n_atoms, n_res=n_atoms, attrs=attrs, atom_resindex=atom_idx
        )
        topo.add_TopologyAttr(Bonds(bond_idx))

        return topo

    @classmethod
    def _read_topo_old(cls, filename) -> Topology:
        """Read topology from old format oxDNA topology file.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        with open(filename) as f:
            first_line = f.readline()

        dimensions = np.array(first_line.split())
        n_atoms = int(dimensions[0])
        array = np.loadtxt(filename, skiprows=1, max_rows=n_atoms, dtype=str)

        # each topology item has two bond columns, which say what the base is bonded
        # from and what it is bonded to. -1 means it is not bonded
        # we need to turn that into an array of bond pairs
        bond_idx = np.zeros((n_atoms * 2, 2), int)
        row_numbers = np.arange(n_atoms, dtype=int)
        for i in range(2):
            rows = row_numbers + (i * n_atoms)
            bond_idx[rows, 0] = array[:, (i + 2)]
            bond_idx[rows, 1] = row_numbers

        # drop any that have -1 as they aren't bonded to anything
        mask = np.logical_and(bond_idx[:, 0] != -1, bond_idx[:, 1] != -1)
        bond_idx = bond_idx[mask, :]

        atom_idx = np.arange(n_atoms, dtype=int)
        attrs = [
            Atomids(atom_idx),
            Resids(atom_idx + 1),
            ChainIDs(array[:, 0].astype(int)),
            Resnames(array[:, 1].astype(str)),
        ]

        topo = Topology(
            n_atoms=n_atoms, n_res=n_atoms, attrs=attrs, atom_resindex=atom_idx
        )
        topo.add_TopologyAttr(Bonds(bond_idx))

        return topo

    def _parseatoms(self):
        """Parse atoms from topology file.

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        if self._is_new_topology(self.filename):
            return self._read_topo_new(self.filename)
        else:
            return self._read_topo_old(self.filename)


def _is_info_line(line: str):
    """
    Check if a line contains simulation information.

    Parameters
    ----------
    line : str
        Line from the OXDNA trajectory file.

    Returns
    -------
    bool
        True if line contains simulation information, False otherwise.
    """
    return line.startswith("t = ") or line.startswith("b = ") or line.startswith("E = ")


class OXDNAReader(ReaderBase):
    """
    Reader for OXDNA trajectory files.

    Parameters
    ----------
    filename : str
        Path to the OXDNA trajectory file.
    **kwargs : dict
        Additional arguments to pass to the reader, must include 'n_atoms'.

    Attributes
    ----------
    n_atoms : int
        Number of atoms in the system.
    ts : Timestep
        Current timestep object.
    n_frames : int
        Total number of frames in trajectory.
    """

    def __init__(self, filename, **kwargs):
        super(OXDNAReader, self).__init__(filename, **kwargs)

        self.n_atoms = kwargs["n_atoms"]
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        oxdnafile = self._oxdnafile = util.anyopen(filename, "rb")

        starts = []
        stops = []

        line = "a new line"
        previous_is_info = False
        current_is_info = False
        while line:
            line = oxdnafile.readline().decode()
            current_is_info = _is_info_line(line)

            # get the lines where the data values start and stop, which
            # are separated by the several lines of overall information
            if not current_is_info and previous_is_info:
                starts.append(oxdnafile.tell() - len(line))
            if current_is_info and not previous_is_info:
                stops.append(oxdnafile.tell() - len(line))

            previous_is_info = current_is_info

        # add the end of the file
        stops += [oxdnafile.tell()]

        self._start_offsets = starts
        self._stop_offsets = stops[1:]  # drop the first
        self.n_frames = len(self._start_offsets)

        self._read_frame(0)

    def _reopen(self):
        """
        Reopen the trajectory file and reset the timestep frame.
        """
        self.close()
        self._oxdnafile = util.anyopen(self.filename, "rb")
        self.ts.frame = -1

    def _read_next_timestep(self, ts=None):
        """
        Read the next timestep from the trajectory.

        Parameters
        ----------
        ts : Timestep, optional
            If provided, update this timestep object instead of creating a new one.

        Returns
        -------
        Timestep
            The timestep object containing the frame data.
        """
        frame = self.frame + 1
        return self._read_frame(frame)

    def _read_frame(self, frame):
        """
        Read a specific frame from the trajectory.

        Parameters
        ----------
        frame : int
            Frame number to read.

        Returns
        -------
        Timestep
            The timestep object containing the frame data.

        Raises
        ------
        OSError
            If the frame number is out of range.
        """
        try:
            start = self._start_offsets[frame]
            stop = self._stop_offsets[frame]
        except IndexError:
            raise OSError from None

        self._oxdnafile.seek(start)
        chunk = self._oxdnafile.read(stop - start)

        array = np.array(
            [
                np.array(line.split(), dtype=float)
                for line in chunk.decode().splitlines()
            ]
        )

        # TODO: also access and update the other values
        self.ts.positions = array[:, :3]

        for i, name in enumerate(
            ("base_vector", "base_normal", "velocity", "angular_velocity")
        ):
            starting_column = 3 * (i + 1)
            if starting_column >= array.shape[1]:
                continue
            cols = np.arange(3) + starting_column
            self.ts.data[name] = array[:, cols]
        self.ts.frame = frame

        return self.ts

    def close(self):
        """
        Close the trajectory file.
        """
        self._oxdnafile.close()


class OXDNA(Trajectory):
    """
    A class to handle oxDNA trajectory data.

    Parameters
    ----------
    universe : MDAnalysis.Universe
        The MDAnalysis Universe object containing the trajectory data
    world_scale : float, optional
        Scaling factor for the world coordinates, by default 0.01

    Attributes
    ----------
    _entity_type : EntityType
        Type of the molecular entity
    _att_names : tuple
        Names of the attributes to track
    """

    def __init__(
        self,
        universe: Universe,
        name: str = "NewOXDNAObject",
        world_scale: float = 0.01,
        create_object: bool = True,
    ):
        self._att_names = (
            "base_vector",
            "base_normal",
            "velocity",
            "angular_velocity",
        )
        super().__init__(
            universe=universe,
            name=name,
            world_scale=world_scale * DNA_SCALE,
            create_object=create_object,
        )

    def _compute_color(self) -> np.ndarray:
        """Compute equidistant chain coloring for OXDNA"""
        return color.color_chains_equidistant(self.named_attribute("chain_id")) / 255

    @property
    def _blender_attributes(self):
        """
        Override parent to provide OXDNA-specific attribute mapping.
        OXDNA stores raw chain IDs and resnames, not the computed indices/numbers.
        """
        return {
            "res_id": self._compute_res_id,
            "chain_id": self._compute_chain_id_int,
            "res_name": self._compute_res_name_int,
            "Color": self._compute_color,
        }

    def _create_object(
        self,
        name: str = "NewUniverseObject",
        style: str | None = "oxdna",
    ):
        """
        Create a new object with the trajectory data.

        Parameters
        ----------
        style : str, optional
            Style of the object representation, by default "oxdna"
        name : str, optional
            Name of the new object, by default "NewUniverseObject"

        Returns
        -------
        bpy.types.Object
            The created Blender object
        """
        self.object = db.create_object(
            name=name,
            collection=coll.mn(),
            vertices=self._scaled_position,
            edges=self.atoms.bonds.indices if hasattr(self.atoms, "bonds") else None,
        )
        self._mn_entity_type = EntityType.MD_OXDNA.value

        self._store_default_attributes()
        self._store_extra_attributes()
        self._update_timestep_values()

        # Setup modifiers
        # self._setup_modifiers()

        # Apply OXDNA-specific style
        if style:
            nodes.create_starting_node_tree(self.object, style="oxdna", color=None)

    def set_frame(self, frame: int) -> None:
        super()._update_positions(frame)
        self._update_timestep_values()

    def _update_timestep_values(self):
        """
        Update the timestep values for all tracked attributes.
        """
        for name in self._att_names:
            try:
                self.store_named_attribute(
                    data=self.universe.trajectory.ts.data[name] * self.world_scale,
                    name=name,
                    atype=db.AttributeTypes.FLOAT_VECTOR,
                )
            except KeyError as e:
                print(e)



import inspect
import types
import typing
import bpy
from PIL import Image


def get_all_class_annotations(cls) -> dict:
    """Get all the annotations from the class including base classes"""
    all_annotations = {}
    for base_cls in cls.mro():
        if hasattr(base_cls, "__annotations__"):
            current_annotations = inspect.get_annotations(base_cls, eval_str=True)
            all_annotations.update(current_annotations)
    if "interface" in all_annotations:
        del all_annotations["interface"]
    return all_annotations


def get_blender_supported_type(atype):
    supported_types = (
        str,
        bool,
        int,
        float,
        tuple[float, float],
        tuple[float, float, float],
        tuple[float, float, float, float],
    )
    if atype in supported_types:
        return atype
    if typing.get_origin(atype) is typing.Union or type(atype) is types.UnionType:
        for utype in atype.__args__:
            if utype in supported_types:
                return utype
    return None


def get_view_matrix(obj):
    if obj._render_mode:
        return obj._scene.camera.matrix_world.inverted()
    else:
        return obj._rv3d.view_matrix


def is_perspective_projection(obj):
    if obj._render_mode:
        return obj._scene.camera.data.type != "ORTHO"
    else:
        return obj._rv3d.is_perspective


def render_annotations(
    scene: bpy.types.Scene, render_scale: float, image: Image, image_scale: float
) -> None:
    """Render annotations of all entities to an image"""
    session = scene.MNSession
    session.prune()  # remove any invalid session entities
    for entity in session.entities.values():
        if not hasattr(entity, "annotations"):
            continue
        manager = entity.annotations
        manager._enable_render_mode(scene, render_scale, image, image_scale)
        manager._draw_annotations()
        manager._disable_render_mode()



import bpy


# initialize style_density_iso_surface node group
def style_density_iso_surface_node_group():
    style_density_iso_surface = bpy.data.node_groups.new(
        type="GeometryNodeTree", name="Style Density ISO Surface"
    )

    # style_density_iso_surface interface
    # Socket Geometry
    geometry_socket = style_density_iso_surface.interface.new_socket(
        name="Geometry", in_out="OUTPUT", socket_type="NodeSocketGeometry"
    )
    geometry_socket.description = "ISO surface geometry output"

    # Socket Volume
    volume_socket = style_density_iso_surface.interface.new_socket(
        name="Volume", in_out="INPUT", socket_type="NodeSocketGeometry"
    )
    volume_socket.description = "Input geometry"

    # Socket Visible
    visible_socket = style_density_iso_surface.interface.new_socket(
        name="Visible", in_out="INPUT", socket_type="NodeSocketBool"
    )
    visible_socket.default_value = True
    visible_socket.description = "Visibility of style"

    # Socket Shade Smooth
    shade_smooth_socket = style_density_iso_surface.interface.new_socket(
        name="Shade Smooth", in_out="INPUT", socket_type="NodeSocketBool"
    )
    shade_smooth_socket.default_value = True
    shade_smooth_socket.description = "Use smooth shading for surface"

    # Socket ISO Value
    iso_value_socket = style_density_iso_surface.interface.new_socket(
        name="ISO Value", in_out="INPUT", socket_type="NodeSocketFloat"
    )
    iso_value_socket.min_value = 0.0
    iso_value_socket.max_value = 1.0
    iso_value_socket.description = "ISO value"

    # Socket Positive Color
    positive_color_socket = style_density_iso_surface.interface.new_socket(
        name="Positive Color", in_out="INPUT", socket_type="NodeSocketColor"
    )
    positive_color_socket.default_value = (0.0, 0.0, 1.0, 1.0)
    positive_color_socket.description = "Color for positive ISO values"

    # Socket Negative Color
    negative_color_socket = style_density_iso_surface.interface.new_socket(
        name="Negative Color", in_out="INPUT", socket_type="NodeSocketColor"
    )
    negative_color_socket.default_value = (1.0, 0.0, 0.0, 1.0)
    negative_color_socket.description = "Color for negative ISO values"

    # Socket Material
    material_socket = style_density_iso_surface.interface.new_socket(
        name="Material", in_out="INPUT", socket_type="NodeSocketMaterial"
    )
    material_socket.description = "Material to use for this surface"

    # Panel Contours
    contours_panel = style_density_iso_surface.interface.new_panel("Contours")
    # Socket Show Contours
    show_contours_socket = style_density_iso_surface.interface.new_socket(
        name="Show Contours",
        in_out="INPUT",
        socket_type="NodeSocketBool",
        parent=contours_panel,
    )
    show_contours_socket.description = "Whether to show surface contours"

    # Socket Only Contours
    only_contours_socket = style_density_iso_surface.interface.new_socket(
        name="Only Contours",
        in_out="INPUT",
        socket_type="NodeSocketBool",
        parent=contours_panel,
    )
    only_contours_socket.description = "Only show contour edges"

    # Socket Contour Thickness
    contour_thickness_socket = style_density_iso_surface.interface.new_socket(
        name="Contour Thickness",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=contours_panel,
    )
    contour_thickness_socket.default_value = 0.1
    contour_thickness_socket.min_value = 0.0
    contour_thickness_socket.max_value = 1.0
    contour_thickness_socket.description = "Thickness of the contour edges"

    # Socket Contour Color
    contour_color_socket = style_density_iso_surface.interface.new_socket(
        name="Contour Color",
        in_out="INPUT",
        socket_type="NodeSocketColor",
        parent=contours_panel,
    )
    contour_color_socket.default_value = (0.0, 0.0, 0.0, 1.0)
    contour_color_socket.description = "Color of contour edges"

    # Panel Slicing
    slicing_panel = style_density_iso_surface.interface.new_panel("Slicing")
    # Socket Slice Left
    slice_left_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Left",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_left_socket.min_value = 0.0
    slice_left_socket.max_value = 100.0
    slice_left_socket.description = "Slice from left (along X axis)"

    # Socket Slice Right
    slice_right_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Right",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_right_socket.min_value = 0.0
    slice_right_socket.max_value = 100.0
    slice_right_socket.description = "Slice from right (along X axis)"

    # Socket Slice Front
    slice_front_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Front",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_front_socket.min_value = 0.0
    slice_front_socket.max_value = 100.0
    slice_front_socket.description = "Slice from front (along Y axis)"

    # Socket Slice Back
    slice_back_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Back",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_back_socket.min_value = 0.0
    slice_back_socket.max_value = 100.0
    slice_back_socket.description = "Slice fom back (along Y axis)"

    # Socket Slice Top
    slice_top_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Top",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_top_socket.min_value = 0.0
    slice_top_socket.max_value = 100.0
    slice_top_socket.description = "Slice from top (along Z axis)"

    # Socket Slice Bottom
    slice_bottom_socket = style_density_iso_surface.interface.new_socket(
        name="Slice Bottom",
        in_out="INPUT",
        socket_type="NodeSocketFloat",
        parent=slicing_panel,
    )
    slice_bottom_socket.min_value = 0.0
    slice_bottom_socket.max_value = 100.0
    slice_bottom_socket.description = "Slice from bottom (along Z axis)"

    # initialize style_density_iso_surface nodes
    # node Group Output
    group_output = style_density_iso_surface.nodes.new("NodeGroupOutput")
    group_output.name = "Group Output"

    # node Group Input
    group_input = style_density_iso_surface.nodes.new("NodeGroupInput")
    group_input.name = "Group Input"

    # node Mesh to Curve
    mesh_to_curve = style_density_iso_surface.nodes.new("GeometryNodeMeshToCurve")
    mesh_to_curve.name = "Mesh to Curve"

    # node Join Geometry Final
    join_geometry_final = style_density_iso_surface.nodes.new(
        "GeometryNodeJoinGeometry"
    )
    join_geometry_final.name = "Join Geometry Final"

    # node Volume to Mesh Negative
    volume_to_mesh_negative = style_density_iso_surface.nodes.new(
        "GeometryNodeVolumeToMesh"
    )
    volume_to_mesh_negative.name = "Volume to Mesh Negative"

    # node Join Geometry Mesh
    join_geometry_mesh = style_density_iso_surface.nodes.new("GeometryNodeJoinGeometry")
    join_geometry_mesh.name = "Join Geometry Mesh"

    # node Compare Y Positive
    compare_y_positive = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_y_positive.name = "Compare Y Positive"
    compare_y_positive.operation = "LESS_EQUAL"

    # node Delete Geometry
    delete_geometry = style_density_iso_surface.nodes.new("GeometryNodeDeleteGeometry")
    delete_geometry.name = "Delete Geometry"

    # node Compare X Positive
    compare_x_positive = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_x_positive.name = "Compare X Positive"
    compare_x_positive.operation = "LESS_EQUAL"

    # node Position
    position = style_density_iso_surface.nodes.new("GeometryNodeInputPosition")
    position.name = "Position"

    # node Separate XYZ
    separate_xyz = style_density_iso_surface.nodes.new("ShaderNodeSeparateXYZ")
    separate_xyz.name = "Separate XYZ"

    # node Boolean Math X
    boolean_math_x = style_density_iso_surface.nodes.new("FunctionNodeBooleanMath")
    boolean_math_x.name = "Boolean Math X"
    boolean_math_x.operation = "OR"

    # node Set Material Positive
    set_material_positive = style_density_iso_surface.nodes.new(
        "GeometryNodeSetMaterial"
    )
    set_material_positive.name = "Set Material Positive"

    # node Set Material Negative
    set_material_negative = style_density_iso_surface.nodes.new(
        "GeometryNodeSetMaterial"
    )
    set_material_negative.name = "Set Material Negative"

    # node Volume to Mesh Positive
    volume_to_mesh_positive = style_density_iso_surface.nodes.new(
        "GeometryNodeVolumeToMesh"
    )
    volume_to_mesh_positive.name = "Volume to Mesh Positive"

    # node Math
    math = style_density_iso_surface.nodes.new("ShaderNodeMath")
    math.name = "Math"
    math.operation = "MULTIPLY"
    # Value_001
    math.inputs[1].default_value = -1.0

    # node Set Shade Smooth
    set_shade_smooth = style_density_iso_surface.nodes.new("GeometryNodeSetShadeSmooth")
    set_shade_smooth.name = "Set Shade Smooth"

    # node Compare Z Positive
    compare_z_positive = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_z_positive.name = "Compare Z Positive"
    compare_z_positive.operation = "LESS_EQUAL"

    # node Boolean Math XY
    boolean_math_xy = style_density_iso_surface.nodes.new("FunctionNodeBooleanMath")
    boolean_math_xy.name = "Boolean Math XY"
    boolean_math_xy.operation = "OR"

    # node Map Range X Positive
    map_range_x_positive = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_x_positive.name = "Map Range X Positive"
    # From Max
    map_range_x_positive.inputs[2].default_value = 100.0

    # node Map Range Y Positive
    map_range_y_positive = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_y_positive.name = "Map Range Y Positive"
    # From Max
    map_range_y_positive.inputs[2].default_value = 100.0

    # node Map Range Z Positive
    map_range_z_positive = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_z_positive.name = "Map Range Z Positive"
    # From Max
    map_range_z_positive.inputs[2].default_value = 100.0

    # node Map Range X Negative
    map_range_x_negative = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_x_negative.name = "Map Range X Negative"
    # From Max
    map_range_x_negative.inputs[2].default_value = 100.0

    # node Compare X Negative
    compare_x_negative = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_x_negative.name = "Compare X Negative"
    compare_x_negative.operation = "GREATER_EQUAL"

    # node Boolean Math XYZ
    boolean_math_xyz = style_density_iso_surface.nodes.new("FunctionNodeBooleanMath")
    boolean_math_xyz.name = "Boolean Math XYZ"
    boolean_math_xyz.operation = "OR"

    # node X Min
    x_min = style_density_iso_surface.nodes.new("ShaderNodeValue")
    x_min.name = "X Min"

    # node X Max
    x_max = style_density_iso_surface.nodes.new("ShaderNodeValue")
    x_max.name = "X Max"

    x_max.outputs["Value"].default_value = 1.0
    # node Frame X
    frame_x = style_density_iso_surface.nodes.new("NodeFrame")
    frame_x.name = "Frame X"

    # node Map Range Y Negative
    map_range_y_negative = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_y_negative.name = "Map Range Y Negative"
    # From Max
    map_range_y_negative.inputs[2].default_value = 100.0

    # node Y Min
    y_min = style_density_iso_surface.nodes.new("ShaderNodeValue")
    y_min.name = "Y Min"

    # node Y Max
    y_max = style_density_iso_surface.nodes.new("ShaderNodeValue")
    y_max.name = "Y Max"

    y_max.outputs["Value"].default_value = 1.0
    # node Compare Y Negative
    compare_y_negative = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_y_negative.name = "Compare Y Negative"
    compare_y_negative.operation = "GREATER_EQUAL"

    # node Boolean Math Y
    boolean_math_y = style_density_iso_surface.nodes.new("FunctionNodeBooleanMath")
    boolean_math_y.name = "Boolean Math Y"
    boolean_math_y.operation = "OR"

    # node Frame Y
    frame_y = style_density_iso_surface.nodes.new("NodeFrame")
    frame_y.name = "Frame Y"

    # node Z Max
    z_max = style_density_iso_surface.nodes.new("ShaderNodeValue")
    z_max.name = "Z Max"

    # node Z Min
    z_min = style_density_iso_surface.nodes.new("ShaderNodeValue")
    z_min.name = "Z Min"

    z_min.outputs["Value"].default_value = 1.0
    # node Map Range Z Negative
    map_range_z_negative = style_density_iso_surface.nodes.new("ShaderNodeMapRange")
    map_range_z_negative.name = "Map Range Z Negative"
    # From Max
    map_range_z_negative.inputs[2].default_value = 100.0

    # node Compare Z Negative
    compare_z_negative = style_density_iso_surface.nodes.new("FunctionNodeCompare")
    compare_z_negative.name = "Compare Z Negative"
    compare_z_negative.operation = "GREATER_EQUAL"

    # node Boolean Math Z
    boolean_math_z = style_density_iso_surface.nodes.new("FunctionNodeBooleanMath")
    boolean_math_z.name = "Boolean Math Z"
    boolean_math_z.operation = "OR"

    # node Frame Z
    frame_z = style_density_iso_surface.nodes.new("NodeFrame")
    frame_z.name = "Frame Z"

    # node Store Named Attribute Positive
    store_named_attribute_positive = style_density_iso_surface.nodes.new(
        "GeometryNodeStoreNamedAttribute"
    )
    store_named_attribute_positive.name = "Store Named Attribute Positive"
    store_named_attribute_positive.data_type = "FLOAT_COLOR"
    # Name
    store_named_attribute_positive.inputs["Name"].default_value = "Color"

    # node Store Named Attribute Negative
    store_named_attribute_negative = style_density_iso_surface.nodes.new(
        "GeometryNodeStoreNamedAttribute"
    )
    store_named_attribute_negative.name = "Store Named Attribute Negative"
    store_named_attribute_negative.data_type = "FLOAT_COLOR"
    # Name
    store_named_attribute_negative.inputs["Name"].default_value = "Color"

    # node Curve to Mesh
    curve_to_mesh = style_density_iso_surface.nodes.new("GeometryNodeCurveToMesh")
    curve_to_mesh.name = "Curve to Mesh"

    # node Quadrilateral
    quadrilateral = style_density_iso_surface.nodes.new(
        "GeometryNodeCurvePrimitiveQuadrilateral"
    )
    quadrilateral.name = "Quadrilateral"

    # node Store Named Attribute Contours
    store_named_attribute_contours = style_density_iso_surface.nodes.new(
        "GeometryNodeStoreNamedAttribute"
    )
    store_named_attribute_contours.name = "Store Named Attribute Contours"
    store_named_attribute_contours.data_type = "FLOAT_COLOR"
    store_named_attribute_contours.domain = "EDGE"
    # Name
    store_named_attribute_contours.inputs["Name"].default_value = "Color"

    # node Set Material Contours
    set_material_contours = style_density_iso_surface.nodes.new(
        "GeometryNodeSetMaterial"
    )
    set_material_contours.name = "Set Material Contours"

    # node Scale Down Contour Thickness
    scale_down_contour_thickness = style_density_iso_surface.nodes.new("ShaderNodeMath")
    scale_down_contour_thickness.name = "Scale Down Contour Thickness"
    scale_down_contour_thickness.operation = "MULTIPLY"
    # Value_001
    scale_down_contour_thickness.inputs[1].default_value = 0.001

    # node Only Contours Switch
    only_contours_switch = style_density_iso_surface.nodes.new("GeometryNodeSwitch")
    only_contours_switch.name = "Only Contours Switch"

    # node Visibility
    visibility = style_density_iso_surface.nodes.new("GeometryNodeSwitch")
    visibility.name = "Visibility"

    # Set parents
    compare_y_positive.parent = frame_y
    compare_x_positive.parent = frame_x
    boolean_math_x.parent = frame_x
    compare_z_positive.parent = frame_z
    map_range_x_positive.parent = frame_x
    map_range_y_positive.parent = frame_y
    map_range_z_positive.parent = frame_z
    map_range_x_negative.parent = frame_x
    compare_x_negative.parent = frame_x
    x_min.parent = frame_x
    x_max.parent = frame_x
    map_range_y_negative.parent = frame_y
    y_min.parent = frame_y
    y_max.parent = frame_y
    compare_y_negative.parent = frame_y
    boolean_math_y.parent = frame_y
    z_max.parent = frame_z
    z_min.parent = frame_z
    map_range_z_negative.parent = frame_z
    compare_z_negative.parent = frame_z
    boolean_math_z.parent = frame_z

    # Set locations
    group_output.location = (2200.0, 40.0)
    group_input.location = (-1100.0, 0.0)
    mesh_to_curve.location = (1180.0, -80.0)
    join_geometry_final.location = (2000.0, 40.0)
    volume_to_mesh_negative.location = (-400.0, 20.0)
    join_geometry_mesh.location = (380.0, 160.0)
    compare_y_positive.location = (550.0, -30.0)
    delete_geometry.location = (1000.0, 20.0)
    compare_x_positive.location = (550.0, -30.0)
    position.location = (-940.0, -700.0)
    separate_xyz.location = (-760.0, -680.0)
    boolean_math_x.location = (730.0, -90.0)
    set_material_positive.location = (-60.0, 220.0)
    set_material_negative.location = (0.0, 40.0)
    volume_to_mesh_positive.location = (-460.0, 200.0)
    math.location = (-580.0, 20.0)
    set_shade_smooth.location = (580.0, 200.0)
    compare_z_positive.location = (530.0, -30.0)
    boolean_math_xy.location = (520.0, -360.0)
    map_range_x_positive.location = (210.0, -70.0)
    map_range_y_positive.location = (210.0, -70.0)
    map_range_z_positive.location = (190.0, -50.0)
    map_range_x_negative.location = (370.0, -70.0)
    compare_x_negative.location = (550.0, -190.0)
    boolean_math_xyz.location = (720.0, -440.0)
    x_min.location = (30.0, -110.0)
    x_max.location = (30.0, -210.0)
    frame_x.location = (-450.0, -190.0)
    map_range_y_negative.location = (370.0, -70.0)
    y_min.location = (30.0, -90.0)
    y_max.location = (30.0, -190.0)
    compare_y_negative.location = (550.0, -190.0)
    boolean_math_y.location = (730.0, -70.0)
    frame_y.location = (-450.0, -590.0)
    z_max.location = (30.0, -170.0)
    z_min.location = (30.0, -70.0)
    map_range_z_negative.location = (350.0, -50.0)
    compare_z_negative.location = (530.0, -190.0)
    boolean_math_z.location = (710.0, -70.0)
    frame_z.location = (-430.0, -990.0)
    store_named_attribute_positive.location = (-240.0, 260.0)
    store_named_attribute_negative.location = (-180.0, 40.0)
    curve_to_mesh.location = (1360.0, -80.0)
    quadrilateral.location = (1180.0, -200.0)
    store_named_attribute_contours.location = (1540.0, -80.0)
    set_material_contours.location = (1720.0, -40.0)
    scale_down_contour_thickness.location = (1020.0, -200.0)
    only_contours_switch.location = (1360.0, 80.0)
    visibility.location = (-840.0, 120.0)

    # Set dimensions
    frame_x.width, frame_x.height = 900.0, 368.0
    frame_y.width, frame_y.height = 900.0, 368.0
    frame_z.width, frame_z.height = 880.0, 368.0

    # initialize style_density_iso_surface links
    # delete_geometry.Geometry -> mesh_to_curve.Mesh
    style_density_iso_surface.links.new(
        delete_geometry.outputs["Geometry"], mesh_to_curve.inputs["Mesh"]
    )
    # separate_xyz.X -> compare_x_positive.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["X"], compare_x_positive.inputs[0]
    )
    # math.Value -> volume_to_mesh_negative.Threshold
    style_density_iso_surface.links.new(
        math.outputs["Value"], volume_to_mesh_negative.inputs["Threshold"]
    )
    # store_named_attribute_positive.Geometry -> set_material_positive.Geometry
    style_density_iso_surface.links.new(
        store_named_attribute_positive.outputs["Geometry"],
        set_material_positive.inputs["Geometry"],
    )
    # join_geometry_mesh.Geometry -> set_shade_smooth.Geometry
    style_density_iso_surface.links.new(
        join_geometry_mesh.outputs["Geometry"], set_shade_smooth.inputs["Geometry"]
    )
    # separate_xyz.Y -> compare_y_positive.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["Y"], compare_y_positive.inputs[0]
    )
    # store_named_attribute_negative.Geometry -> set_material_negative.Geometry
    style_density_iso_surface.links.new(
        store_named_attribute_negative.outputs["Geometry"],
        set_material_negative.inputs["Geometry"],
    )
    # set_shade_smooth.Geometry -> delete_geometry.Geometry
    style_density_iso_surface.links.new(
        set_shade_smooth.outputs["Geometry"], delete_geometry.inputs["Geometry"]
    )
    # position.Position -> separate_xyz.Vector
    style_density_iso_surface.links.new(
        position.outputs["Position"], separate_xyz.inputs["Vector"]
    )
    # set_material_negative.Geometry -> join_geometry_mesh.Geometry
    style_density_iso_surface.links.new(
        set_material_negative.outputs["Geometry"], join_geometry_mesh.inputs["Geometry"]
    )
    # compare_x_positive.Result -> boolean_math_x.Boolean
    style_density_iso_surface.links.new(
        compare_x_positive.outputs["Result"], boolean_math_x.inputs[0]
    )
    # group_input.ISO Value -> volume_to_mesh_positive.Threshold
    style_density_iso_surface.links.new(
        group_input.outputs["ISO Value"], volume_to_mesh_positive.inputs["Threshold"]
    )
    # group_input.ISO Value -> math.Value
    style_density_iso_surface.links.new(
        group_input.outputs["ISO Value"], math.inputs[0]
    )
    # group_input.Show Contours -> mesh_to_curve.Selection
    style_density_iso_surface.links.new(
        group_input.outputs["Show Contours"], mesh_to_curve.inputs["Selection"]
    )
    # visibility.Output -> volume_to_mesh_positive.Volume
    style_density_iso_surface.links.new(
        visibility.outputs["Output"], volume_to_mesh_positive.inputs["Volume"]
    )
    # join_geometry_final.Geometry -> group_output.Geometry
    style_density_iso_surface.links.new(
        join_geometry_final.outputs["Geometry"], group_output.inputs["Geometry"]
    )
    # boolean_math_x.Boolean -> boolean_math_xy.Boolean
    style_density_iso_surface.links.new(
        boolean_math_x.outputs["Boolean"], boolean_math_xy.inputs[0]
    )
    # group_input.Slice Left -> map_range_x_positive.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Left"], map_range_x_positive.inputs["Value"]
    )
    # group_input.Slice Front -> map_range_y_positive.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Front"], map_range_y_positive.inputs["Value"]
    )
    # group_input.Slice Top -> map_range_z_negative.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Top"], map_range_z_negative.inputs["Value"]
    )
    # map_range_y_positive.Result -> compare_y_positive.B
    style_density_iso_surface.links.new(
        map_range_y_positive.outputs["Result"], compare_y_positive.inputs[1]
    )
    # separate_xyz.Z -> compare_z_positive.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["Z"], compare_z_positive.inputs[0]
    )
    # map_range_z_positive.Result -> compare_z_positive.B
    style_density_iso_surface.links.new(
        map_range_z_positive.outputs["Result"], compare_z_positive.inputs[1]
    )
    # map_range_x_positive.Result -> compare_x_positive.B
    style_density_iso_surface.links.new(
        map_range_x_positive.outputs["Result"], compare_x_positive.inputs[1]
    )
    # group_input.Slice Right -> map_range_x_negative.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Right"], map_range_x_negative.inputs["Value"]
    )
    # separate_xyz.X -> compare_x_negative.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["X"], compare_x_negative.inputs[0]
    )
    # map_range_x_negative.Result -> compare_x_negative.B
    style_density_iso_surface.links.new(
        map_range_x_negative.outputs["Result"], compare_x_negative.inputs[1]
    )
    # compare_x_negative.Result -> boolean_math_x.Boolean
    style_density_iso_surface.links.new(
        compare_x_negative.outputs["Result"], boolean_math_x.inputs[1]
    )
    # boolean_math_z.Boolean -> boolean_math_xyz.Boolean
    style_density_iso_surface.links.new(
        boolean_math_z.outputs["Boolean"], boolean_math_xyz.inputs[1]
    )
    # x_min.Value -> map_range_x_positive.To Min
    style_density_iso_surface.links.new(
        x_min.outputs["Value"], map_range_x_positive.inputs[3]
    )
    # x_max.Value -> map_range_x_positive.To Max
    style_density_iso_surface.links.new(
        x_max.outputs["Value"], map_range_x_positive.inputs[4]
    )
    # x_min.Value -> map_range_x_negative.To Max
    style_density_iso_surface.links.new(
        x_min.outputs["Value"], map_range_x_negative.inputs[4]
    )
    # x_max.Value -> map_range_x_negative.To Min
    style_density_iso_surface.links.new(
        x_max.outputs["Value"], map_range_x_negative.inputs[3]
    )
    # y_min.Value -> map_range_y_positive.To Min
    style_density_iso_surface.links.new(
        y_min.outputs["Value"], map_range_y_positive.inputs[3]
    )
    # y_max.Value -> map_range_y_positive.To Max
    style_density_iso_surface.links.new(
        y_max.outputs["Value"], map_range_y_positive.inputs[4]
    )
    # y_min.Value -> map_range_y_negative.To Max
    style_density_iso_surface.links.new(
        y_min.outputs["Value"], map_range_y_negative.inputs[4]
    )
    # y_max.Value -> map_range_y_negative.To Min
    style_density_iso_surface.links.new(
        y_max.outputs["Value"], map_range_y_negative.inputs[3]
    )
    # separate_xyz.Y -> compare_y_negative.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["Y"], compare_y_negative.inputs[0]
    )
    # map_range_y_negative.Result -> compare_y_negative.B
    style_density_iso_surface.links.new(
        map_range_y_negative.outputs["Result"], compare_y_negative.inputs[1]
    )
    # compare_y_positive.Result -> boolean_math_y.Boolean
    style_density_iso_surface.links.new(
        compare_y_positive.outputs["Result"], boolean_math_y.inputs[0]
    )
    # compare_y_negative.Result -> boolean_math_y.Boolean
    style_density_iso_surface.links.new(
        compare_y_negative.outputs["Result"], boolean_math_y.inputs[1]
    )
    # group_input.Slice Back -> map_range_y_negative.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Back"], map_range_y_negative.inputs["Value"]
    )
    # z_min.Value -> map_range_z_positive.To Min
    style_density_iso_surface.links.new(
        z_min.outputs["Value"], map_range_z_positive.inputs[3]
    )
    # z_max.Value -> map_range_z_positive.To Max
    style_density_iso_surface.links.new(
        z_max.outputs["Value"], map_range_z_positive.inputs[4]
    )
    # z_min.Value -> map_range_z_negative.To Max
    style_density_iso_surface.links.new(
        z_min.outputs["Value"], map_range_z_negative.inputs[4]
    )
    # z_max.Value -> map_range_z_negative.To Min
    style_density_iso_surface.links.new(
        z_max.outputs["Value"], map_range_z_negative.inputs[3]
    )
    # group_input.Slice Bottom -> map_range_z_positive.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Slice Bottom"], map_range_z_positive.inputs["Value"]
    )
    # map_range_z_negative.Result -> compare_z_negative.B
    style_density_iso_surface.links.new(
        map_range_z_negative.outputs["Result"], compare_z_negative.inputs[1]
    )
    # separate_xyz.Z -> compare_z_negative.A
    style_density_iso_surface.links.new(
        separate_xyz.outputs["Z"], compare_z_negative.inputs[0]
    )
    # compare_z_positive.Result -> boolean_math_z.Boolean
    style_density_iso_surface.links.new(
        compare_z_positive.outputs["Result"], boolean_math_z.inputs[0]
    )
    # compare_z_negative.Result -> boolean_math_z.Boolean
    style_density_iso_surface.links.new(
        compare_z_negative.outputs["Result"], boolean_math_z.inputs[1]
    )
    # boolean_math_y.Boolean -> boolean_math_xy.Boolean
    style_density_iso_surface.links.new(
        boolean_math_y.outputs["Boolean"], boolean_math_xy.inputs[1]
    )
    # boolean_math_xy.Boolean -> boolean_math_xyz.Boolean
    style_density_iso_surface.links.new(
        boolean_math_xy.outputs["Boolean"], boolean_math_xyz.inputs[0]
    )
    # boolean_math_xyz.Boolean -> delete_geometry.Selection
    style_density_iso_surface.links.new(
        boolean_math_xyz.outputs["Boolean"], delete_geometry.inputs["Selection"]
    )
    # group_input.Shade Smooth -> set_shade_smooth.Shade Smooth
    style_density_iso_surface.links.new(
        group_input.outputs["Shade Smooth"], set_shade_smooth.inputs["Shade Smooth"]
    )
    # volume_to_mesh_positive.Mesh -> store_named_attribute_positive.Geometry
    style_density_iso_surface.links.new(
        volume_to_mesh_positive.outputs["Mesh"],
        store_named_attribute_positive.inputs["Geometry"],
    )
    # volume_to_mesh_negative.Mesh -> store_named_attribute_negative.Geometry
    style_density_iso_surface.links.new(
        volume_to_mesh_negative.outputs["Mesh"],
        store_named_attribute_negative.inputs["Geometry"],
    )
    # group_input.Positive Color -> store_named_attribute_positive.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Positive Color"],
        store_named_attribute_positive.inputs["Value"],
    )
    # group_input.Negative Color -> store_named_attribute_negative.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Negative Color"],
        store_named_attribute_negative.inputs["Value"],
    )
    # group_input.Material -> set_material_positive.Material
    style_density_iso_surface.links.new(
        group_input.outputs["Material"], set_material_positive.inputs["Material"]
    )
    # group_input.Material -> set_material_negative.Material
    style_density_iso_surface.links.new(
        group_input.outputs["Material"], set_material_negative.inputs["Material"]
    )
    # set_material_contours.Geometry -> join_geometry_final.Geometry
    style_density_iso_surface.links.new(
        set_material_contours.outputs["Geometry"],
        join_geometry_final.inputs["Geometry"],
    )
    # mesh_to_curve.Curve -> curve_to_mesh.Curve
    style_density_iso_surface.links.new(
        mesh_to_curve.outputs["Curve"], curve_to_mesh.inputs["Curve"]
    )
    # quadrilateral.Curve -> curve_to_mesh.Profile Curve
    style_density_iso_surface.links.new(
        quadrilateral.outputs["Curve"], curve_to_mesh.inputs["Profile Curve"]
    )
    # curve_to_mesh.Mesh -> store_named_attribute_contours.Geometry
    style_density_iso_surface.links.new(
        curve_to_mesh.outputs["Mesh"], store_named_attribute_contours.inputs["Geometry"]
    )
    # store_named_attribute_contours.Geometry -> set_material_contours.Geometry
    style_density_iso_surface.links.new(
        store_named_attribute_contours.outputs["Geometry"],
        set_material_contours.inputs["Geometry"],
    )
    # group_input.Material -> set_material_contours.Material
    style_density_iso_surface.links.new(
        group_input.outputs["Material"], set_material_contours.inputs["Material"]
    )
    # scale_down_contour_thickness.Value -> quadrilateral.Width
    style_density_iso_surface.links.new(
        scale_down_contour_thickness.outputs["Value"], quadrilateral.inputs["Width"]
    )
    # scale_down_contour_thickness.Value -> quadrilateral.Height
    style_density_iso_surface.links.new(
        scale_down_contour_thickness.outputs["Value"], quadrilateral.inputs["Height"]
    )
    # group_input.Contour Thickness -> scale_down_contour_thickness.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Contour Thickness"], scale_down_contour_thickness.inputs[0]
    )
    # delete_geometry.Geometry -> only_contours_switch.False
    style_density_iso_surface.links.new(
        delete_geometry.outputs["Geometry"], only_contours_switch.inputs["False"]
    )
    # group_input.Only Contours -> only_contours_switch.Switch
    style_density_iso_surface.links.new(
        group_input.outputs["Only Contours"], only_contours_switch.inputs["Switch"]
    )
    # group_input.Contour Color -> store_named_attribute_contours.Value
    style_density_iso_surface.links.new(
        group_input.outputs["Contour Color"],
        store_named_attribute_contours.inputs["Value"],
    )
    # group_input.Volume -> visibility.True
    style_density_iso_surface.links.new(
        group_input.outputs["Volume"], visibility.inputs["True"]
    )
    # visibility.Output -> volume_to_mesh_negative.Volume
    style_density_iso_surface.links.new(
        visibility.outputs["Output"], volume_to_mesh_negative.inputs["Volume"]
    )
    # group_input.Visible -> visibility.Switch
    style_density_iso_surface.links.new(
        group_input.outputs["Visible"], visibility.inputs["Switch"]
    )
    # set_material_positive.Geometry -> join_geometry_mesh.Geometry
    style_density_iso_surface.links.new(
        set_material_positive.outputs["Geometry"], join_geometry_mesh.inputs["Geometry"]
    )
    # only_contours_switch.Output -> join_geometry_final.Geometry
    style_density_iso_surface.links.new(
        only_contours_switch.outputs["Output"], join_geometry_final.inputs["Geometry"]
    )
    return style_density_iso_surface



import gzip
import shutil
from pathlib import Path
import pytest
import molecularnodes as mn
from molecularnodes.nodes import nodes
from .constants import data_dir

cellpack_dir = data_dir / "cellpack/petworld"


files_to_test = [f for f in cellpack_dir.glob("*") if str(f).endswith((".gz", ".bcif"))]


def maybe_unzip(file):
    if str(file).endswith(".gz"):
        # Create a temporary unzipped file
        unzipped_path = str(file)[:-3]  # Remove .gz extension
        if not Path(unzipped_path).exists():
            with gzip.open(file, "rb") as f_in:
                with open(unzipped_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
        return unzipped_path
    else:
        return file


@pytest.mark.parametrize("file", files_to_test)
def test_load_petworld(file):
    file_path = maybe_unzip(data_dir / "cellpack" / file)

    _ens = mn.entities.ensemble.load_cellpack(
        file_path=file_path,
        name="CellPack",
        node_setup=False,
        fraction=0.1,
    )


@pytest.mark.parametrize("format", ["bcif", "cif"])
def test_load_cellpack(snapshot, format):
    file_path = data_dir / f"cellpack/square1.{format}"

    ens = mn.entities.ensemble.load_cellpack(file_path, node_setup=False, fraction=0.1)
    assert ens._entity_type == mn.entities.base.EntityType.ENSEMBLE_CELLPACK
    assert ens.object.mn.entity_type == ens._entity_type.value

    assert ens.name == Path(file_path).name
    assert snapshot == str(ens.object["chain_ids"])
    obj_names = [obj.name for obj in ens.instance_collection.objects]
    assert snapshot == "\n".join(obj_names)

    ens.node_group.nodes["Ensemble Instance"].inputs["As Points"].default_value = False
    nodes.realize_instances(ens.object)
    for attribute in ens.list_attributes():
        assert snapshot == ens[attribute]

    pos_eval = ens.named_attribute("position", evaluate=True)
    assert snapshot == pos_eval.shape
    assert snapshot == pos_eval

    _obj = ens.object



from . import selections
from .annotations import TrajectoryAnnotation, TrajectoryAnnotationManager
from .base import Trajectory
from .io import load, load_oxdna
from .oxdna import OXDNA

__all__ = [
    "selections",
    "Trajectory",
    "load",
    "load_oxdna",
    "OXDNA",
    "TrajectoryAnnotation",
    "TrajectoryAnnotationManager",
]



from pathlib import Path
import numpy as np
from biotite import InvalidFileError
from biotite import structure as struc
from biotite.structure.io import pdbx
from ...utils import array_quaternions_from_dict
from ..molecule.pdbx import PDBXReader


# For reading cellpack files, we override the CIFFile from biotite. The only change we
# implement is that we add the `line.strip()` when initially getting the lines from the
# text in the `deserialize` method. This fixes the reading of the PETWORLD files, and we
# don't have to write-out a modifier version of the file before reading it back in
class CellPackCIFFileReader(pdbx.CIFFile):
    @classmethod
    def read(cls, file):
        # File name
        if pdbx.cif.is_open_compatible(file):
            with open(file, "r") as f:
                text = f.read()
        # File object
        else:
            if not pdbx.cif.is_text(file):
                raise TypeError("A file opened in 'text' mode is required")
            text = file.read()
        return cls.deserialize(text)

    @staticmethod
    def deserialize(text):
        lines = [line.strip() for line in text.splitlines()]
        block_starts = []
        block_names = []
        for i, line in enumerate(lines):
            if not pdbx.cif._is_empty(line):
                data_block_name = pdbx.cif._parse_data_block_name(line)
                if data_block_name is not None:
                    block_starts.append(i)
                    block_names.append(data_block_name)
        return pdbx.CIFFile(
            pdbx.cif._create_element_dict(lines, block_names, block_starts)
        )


class CellPackReader(PDBXReader):
    def __init__(self, file_path):
        self._extra_annotations = {
            "sec_struct": self._get_secondary_structure,
            "entity_id": self._get_entity_id,
            "asym_id": self._get_asym_id,
            "pdb_model_num": self._get_pdb_model_num,
        }
        self._extra_fields = ["b_factor", "occupancy", "atom_id"]
        self.file_path = file_path
        self.file = self.read(file_path)
        self.n_molecules: int = pdbx.get_model_count(self.file)
        self.molecules: dict[str, struc.AtomArray] = self.get_molecules()

    @property
    def mol_ids(self) -> np.ndarray:
        return np.unique(list(self.molecules.keys()))

    def read(self, file_path):
        suffix = Path(file_path).suffix
        if suffix == ".bcif":
            return pdbx.BinaryCIFFile.read(self.file_path)
        elif suffix == ".cif":
            return CellPackCIFFileReader.read(self.file_path)
        else:
            raise ValueError(f"Invalid file format: '{suffix}")

    @staticmethod
    def _get_asym_id(array, file) -> np.ndarray:
        return array.chain_id

    @staticmethod
    def _get_pdb_model_num(array, file) -> np.ndarray:
        return (
            list(file.values())[0]["atom_site"]["pdbx_PDB_model_num"]
            .as_array()
            .astype(int)
        )

    @property
    def blocks(self):
        return list(self.file.values())[0]

    def get_assemblies(self):
        return array_quaternions_from_dict(self._assemblies())

    def get_molecules(self) -> dict[str, struc.AtomArray]:
        self._is_petworld = False

        if "PDB_model_num" in self.blocks["pdbx_struct_assembly_gen"]:
            self._is_petworld = True

        molecules = {}

        try:
            # Skip bond computation for CellPack files as it's extremely slow with large structures
            # (biotite 1.5+ regression: connect_via_residue_names takes ~80s for 200k atoms)
            array = self.get_structure(include_bonds=False)

            if isinstance(array, struc.AtomArrayStack):
                array = array[0]

            molecules = {
                c: array[array.chain_id == c] for c in np.unique(array.chain_id)
            }

        except InvalidFileError:
            self._is_petworld = True
            for i in range(self.n_molecules):
                # Skip bond computation for CellPack files
                array = self.get_structure(model=int(i + 1), include_bonds=False)
                array.set_annotation("pdbx_PDB_model_num", np.repeat(i + 1, len(array)))
                array.chain_id = array.pdbx_PDB_model_num
                chain_name = "{}_{}".format(
                    str(i).rjust(4, "0"), str(array.chain_id[0])
                )
                molecules[chain_name] = array

        for name, array in molecules.items():
            array = self.set_extra_annotations(
                self._extra_annotations, array, self.file
            )
            array = self.set_standard_annotations(array)
            molecules[name] = array

        return molecules



import itertools
import shutil
from pathlib import Path
import bpy

SUBFOLDER = "Molecular Nodes"


def list_templates():
    paths = bpy.utils.app_template_paths()
    t_paths = list(paths) if paths is not None else []
    names = list(
        itertools.chain.from_iterable(
            [[folder.stem for folder in Path(p).rglob("")] for p in t_paths]
        )
    )
    return [name for name in names if not name.startswith("bl_app")]


def is_installed():
    base_path = Path(
        bpy.utils.user_resource(
            "SCRIPTS", path=str(Path("startup") / "bl_app_templates_user"), create=False
        )
    )
    molecular_nodes_path = base_path / "Molecular Nodes"
    return molecular_nodes_path.exists()


def install():
    base_path = Path(
        bpy.utils.user_resource(
            "SCRIPTS", path=str(Path("startup") / "bl_app_templates_user"), create=True
        )
    )

    base_path.mkdir(parents=True, exist_ok=True)
    path_app_templates = base_path / SUBFOLDER
    path_app_templates.mkdir(parents=True, exist_ok=True)
    startup_file = Path(__file__).parent / "template/startup.blend"
    shutil.copy(startup_file, path_app_templates)
    bpy.utils.refresh_script_paths()


def uninstall():
    for folder in bpy.utils.app_template_paths():
        path = Path(folder).absolute() / SUBFOLDER
        if "Molecular Nodes" not in str(path):
            continue

        if path.exists():
            shutil.rmtree(path)
    bpy.utils.refresh_script_paths()



import os
from pathlib import Path

codes = ["4ozs", "8H1B", "1BNA", "8U8W"]
attributes = [
    "b_factor",
    "occupancy",
    "vdw_radii",
    "lipophobicity",
    "charge",
    "res_id",
    "res_name",
    "atomic_number",
    "chain_id",
    "entity_id",
    "atom_id",
    "atom_name",
    "sec_struct",
    "Color",
    "position",
    "is_backbone",
    "is_side_chain",
    "is_alpha_carbon",
    "is_solvent",
    "is_nucleic",
    "is_peptide",
    "is_hetero",
    "is_carb",
    "bond_type",
    "mass",
    "ures_id",
]
data_dir = Path(os.path.abspath(Path(__file__).parent / "data"))



import bpy
import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis.transformations import boxdimensions
import molecularnodes as mn
from .constants import data_dir

bpy.utils.expose_bundled_modules()


class TestAnnotations:
    @pytest.fixture(scope="module")
    def universe(self):
        topo = data_dir / "md_ppr/box.gro"
        traj = data_dir / "md_ppr/first_5_frames.xtc"
        return mda.Universe(topo, traj)

    @pytest.fixture(scope="module")
    def density_file(self):
        file = data_dir / "emd_24805.map.gz"
        vdb_file = data_dir / "emd_24805.vdb"
        vdb_file.unlink(missing_ok=True)
        # Make all densities are removed
        for o in bpy.data.objects:
            if o.mn.entity_type == "density":
                bpy.data.objects.remove(o, do_unlink=True)
        return file

    def test_registered_trajectory_annotations(self):
        # test if trajectory annotations are correctly auto registered
        manager = mn.entities.trajectory.TrajectoryAnnotationManager
        # atom_info
        assert "atom_info" in manager._classes
        assert hasattr(manager, "add_atom_info")
        assert callable(getattr(manager, "add_atom_info"))
        # com
        assert "com" in manager._classes
        assert hasattr(manager, "add_com")
        assert callable(getattr(manager, "add_com"))
        # com_distance
        assert "com_distance" in manager._classes
        assert hasattr(manager, "add_com_distance")
        assert callable(getattr(manager, "add_com_distance"))
        # canonical_dihedrals
        assert "canonical_dihedrals" in manager._classes
        assert hasattr(manager, "add_canonical_dihedrals")
        assert callable(getattr(manager, "add_canonical_dihedrals"))
        # universe_info
        assert "universe_info" in manager._classes
        assert hasattr(manager, "add_universe_info")
        assert callable(getattr(manager, "add_universe_info"))
        # label_2d
        assert "label_2d" in manager._classes
        assert hasattr(manager, "add_label_2d")
        assert callable(getattr(manager, "add_label_2d"))
        # label_3d
        assert "label_3d" in manager._classes
        assert hasattr(manager, "add_label_3d")
        assert callable(getattr(manager, "add_label_3d"))

    def test_registered_molecule_annotations(self):
        # test if molecule annotations are correctly auto registered
        manager = mn.entities.molecule.annotations.MoleculeAnnotationManager
        # molecule_info
        assert "molecule_info" in manager._classes
        assert hasattr(manager, "add_molecule_info")
        assert callable(getattr(manager, "add_molecule_info"))
        # label_2d
        assert "label_2d" in manager._classes
        assert hasattr(manager, "add_label_2d")
        assert callable(getattr(manager, "add_label_2d"))
        # label_3d
        assert "label_3d" in manager._classes
        assert hasattr(manager, "add_label_3d")
        assert callable(getattr(manager, "add_label_3d"))

    def test_registered_density_annotations(self):
        # test if density annotations are correctly auto registered
        manager = mn.entities.density.annotations.DensityAnnotationManager
        # density_info
        assert "density_info" in manager._classes
        assert hasattr(manager, "add_density_info")
        assert callable(getattr(manager, "add_density_info"))
        # grid_axes
        assert "grid_axes" in manager._classes
        assert hasattr(manager, "add_grid_axes")
        assert callable(getattr(manager, "add_grid_axes"))
        # label_2d
        assert "label_2d" in manager._classes
        assert hasattr(manager, "add_label_2d")
        assert callable(getattr(manager, "add_label_2d"))
        # label_3d
        assert "label_3d" in manager._classes
        assert hasattr(manager, "add_label_3d")
        assert callable(getattr(manager, "add_label_3d"))

    def test_trajectory_annotations_registration(self, universe):
        manager = mn.entities.trajectory.TrajectoryAnnotationManager

        # test register exceptions
        class TestAnnotation:
            pass

        # test presence of annotation_type
        with pytest.raises(ValueError):
            manager.register(TestAnnotation)

        class TestAnnotation:
            annotation_type = "test_annotation"

        # test not an annotation class
        with pytest.raises(ValueError):
            manager.register(TestAnnotation)

        # test no draw method
        with pytest.raises(ValueError):

            class TestAnnotation(mn.entities.trajectory.TrajectoryAnnotation):
                annotation_type = "test_annotation"

        class TestAnnotation:
            annotation_type = "test_annotation"

        # test unregister exceptions
        with pytest.raises(ValueError):
            manager.unregister(TestAnnotation)

        class TestAnnotation(mn.entities.trajectory.TrajectoryAnnotation):
            annotation_type = "test_annotation"

            selection: str  # required param
            param1: int = 0
            param2: float = 0.0

            def draw(self):
                print(
                    self.trajectory,
                    self.interface.selection,
                    self.interface.param1,
                    self.interface.param2,
                )

        # test auto register
        assert "test_annotation" in manager._classes
        assert hasattr(manager, "add_test_annotation")
        # test unregister
        manager.unregister(TestAnnotation)
        assert "test_annotation" not in manager._classes
        assert not hasattr(manager, "add_test_annotation")
        # test register
        manager.register(TestAnnotation)
        assert "test_annotation" in manager._classes
        assert hasattr(manager, "add_test_annotation")
        # test add
        t1 = mn.Trajectory(universe)
        # test add with required param missing
        with pytest.raises(ValueError):
            a1 = t1.annotations.add_test_annotation()
        # test add with invalid param
        with pytest.raises(ValueError):
            a1 = t1.annotations.add_test_annotation(invalid_param="value")
        a1 = t1.annotations.add_test_annotation(selection="all")
        # test draw
        a1._instance.draw()

    def test_trajectory_annotation_lifecycle(self, universe):
        t1 = mn.Trajectory(universe)
        # test annotation atom_info add
        assert len(t1.annotations._interfaces) == 0
        a1 = t1.annotations.add_atom_info(selection="all")
        assert len(t1.annotations._interfaces) == 1
        # test len and iterable access
        assert len(t1.annotations) == 1
        for ant in t1.annotations:
            assert ant == a1
        # test subscriptable access
        assert t1.annotations["Annotation"] == a1
        # test index based access
        assert t1.annotations[0] == a1
        # test invalid index
        with pytest.raises(ValueError):
            a1 = t1.annotations[100]
        # test annotation get by name
        ga1 = t1.annotations.get("Annotation")
        assert ga1 == a1
        # test invalid get by name
        with pytest.raises(ValueError):
            ga1 = t1.annotations.get("InvalidAnnotation")
        # test key completions
        list1 = t1.annotations._ipython_key_completions_()
        assert list1 == ["Annotation"]
        # test annotation inputs in interface
        assert hasattr(a1, "selection")
        assert hasattr(a1, "show_resid")
        assert hasattr(a1, "show_segid")
        # test annotation common params in interface
        assert hasattr(a1, "visible")
        # test annotation instance having trajectory entity
        assert a1._instance.trajectory == t1
        # test annotation instance interface
        assert a1 == a1._instance.interface
        # test all annotations visibility
        assert t1.annotations.visible
        t1.annotations.visible = False
        assert not t1.annotations.visible
        # test remove annotation
        a2 = t1.annotations.add_atom_info(selection="resid 2", name="A2")
        a3 = t1.annotations.add_atom_info(selection="resid 3")
        assert a2 in t1.annotations
        assert a3 in t1.annotations
        # remove annotation by name
        t1.annotations.remove("A2")
        assert a2 not in t1.annotations
        # remove annotation by instance
        t1.annotations.remove(a3)
        assert a3 not in t1.annotations
        # test invalid remove by name
        with pytest.raises(ValueError):
            t1.annotations.remove("InvalidAnnotation")
        # test invalid remove by instance
        with pytest.raises(ValueError):
            t1.annotations.remove(int)
        # test invalid remove by instance uuid
        orig_uuid = a1._uuid
        a1._uuid = "InvalidUUID"
        with pytest.raises(ValueError):
            t1.annotations.remove(a1)
        a1._uuid = orig_uuid
        # test clear annotations (remove all)
        assert len(t1.annotations) != 0
        t1.annotations.clear()
        assert len(t1.annotations) == 0
        # test invalid remove by uuid
        with pytest.raises(ValueError):
            t1.annotations._remove_annotation_by_uuid("InvalidUUID")

    def test_annotation_ops(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        # add annotation operator
        # first annotation type due to no invoke - must have no required inputs
        bpy.ops.mn.add_annotation("EXEC_DEFAULT", uuid=t1.uuid)
        assert len(t1.annotations) == 1
        a1 = t1.annotations[0]
        # remove annotation operator
        bpy.ops.mn.remove_annotation(
            "EXEC_DEFAULT", uuid=t1.uuid, annotation_uuid=a1._uuid
        )
        assert len(t1.annotations) == 0

    def test_trajectory_annotation_atom_info(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        # test defaults
        t1.annotations.add_atom_info()
        assert len(t1.annotations) == 1
        t1.annotations.clear()
        # test selection string
        phrase = "all"
        a1 = t1.annotations.add_atom_info(selection=phrase)
        assert len(t1.annotations) == 1
        assert phrase == a1.selection
        t1.annotations.clear()
        # test selection atom group
        ag = universe.select_atoms("all")
        a1 = t1.annotations.add_atom_info(selection=ag)
        assert len(t1.annotations) == 1
        assert ag == a1.selection
        t1.annotations.clear()
        # test invalid selection type (not str or AtomGroup)
        with pytest.raises(ValueError):
            t1.annotations.add_atom_info(selection=1)

    def test_trajectory_annotation_com(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        # test defaults
        t1.annotations.add_com()
        assert len(t1.annotations) == 1
        t1.annotations.clear()
        # test selection string
        phrase = "all"
        a1 = t1.annotations.add_com(selection=phrase)
        assert len(t1.annotations) == 1
        assert phrase == a1.selection
        t1.annotations.clear()
        # test selection atom group
        ag = universe.select_atoms("all")
        a1 = t1.annotations.add_com(selection=ag)
        assert len(t1.annotations) == 1
        assert ag == a1.selection
        t1.annotations.clear()
        # test invalid selection type (not str or AtomGroup)
        with pytest.raises(ValueError):
            t1.annotations.add_com(selection=1)

    def test_trajectory_annotation_com_distance(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        # test defaults
        with pytest.raises(ValueError):
            t1.annotations.add_com_distance()
        assert len(t1.annotations) == 0
        # test both required selections
        with pytest.raises(ValueError):
            t1.annotations.add_com_distance(selection1="all")
        assert len(t1.annotations) == 0
        # test selection strings
        phrase1 = "all"
        a1 = t1.annotations.add_com_distance(selection1=phrase1, selection2="protein")
        assert len(t1.annotations) == 1
        assert phrase1 == a1.selection1
        t1.annotations.clear()
        # test selection atom groups
        ag1 = universe.select_atoms("all")
        a1 = t1.annotations.add_com_distance(selection1=ag1, selection2="protein")
        assert len(t1.annotations) == 1
        assert ag1 == a1.selection1
        t1.annotations.clear()
        # test invalid selection type (not str or AtomGroup)
        with pytest.raises(ValueError):
            t1.annotations.add_com_distance(selection1=1, selection2=2)

    def test_trajectory_annotation_canonical_dihedrals(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        # test defaults - needs resid input
        with pytest.raises(ValueError):
            t1.annotations.add_canonical_dihedrals()
        assert len(t1.annotations) == 0
        # test invalid resid
        with pytest.raises(IndexError):
            t1.annotations.add_canonical_dihedrals(resid=1000)
        assert len(t1.annotations) == 0
        # test valid resid
        a1 = t1.annotations.add_canonical_dihedrals(resid=1)
        assert len(t1.annotations) == 1
        # test change of resid through API
        a1.resid = 2

    def test_trajectory_annotation_universe_info(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        t1.annotations.add_universe_info()
        assert len(t1.annotations) == 1

    def test_trajectory_annotation_simulation_box(self, universe):
        universe_copy = universe.copy()
        t1 = mn.Trajectory(universe_copy, name="TestUniverse")
        assert len(t1.annotations) == 0
        box = t1.annotations.add_simulation_box()
        assert len(t1.annotations) == 1
        obj_name = "MN_an_TestUniverse"
        assert obj_name in bpy.data.objects
        obj_data = bpy.data.objects[obj_name].data
        # verify default triclinic cell in wireframe mode
        vef = [len(obj_data.vertices), len(obj_data.edges), len(obj_data.polygons)]
        assert vef == [8, 12, 0]
        # verify default triclinic cell faces in non-wireframe mode
        box.mesh_wireframe = False
        vef = [len(obj_data.vertices), len(obj_data.edges), len(obj_data.polygons)]
        assert vef == [8, 12, 6]
        # verify common compact shapes
        box_dimensions = np.array(
            [
                [10, 10, 10, 90, 90, 90],  # cubic
                [10, 10, 10, 60, 60, 90],  # rhombic dodecahedron xy-square
                [10, 10, 10, 60, 60, 60],  # rhombic dodecahedron xy-hexagon
                [10, 10, 10, 70.53, 109.47, 70.53],  # truncated octahedron
                [10, 10, 20, 90, 90, 60],  # hexagonal cell
            ]
        )
        # add mda transform to set different box vectors for different frames
        transform = boxdimensions.set_dimensions(box_dimensions)
        universe_copy.trajectory.add_transformations(transform)
        # enable compact option for annotation
        box.compact = True
        # expected # of vertices, edges and faces for standard compact shapes
        expected_vef = [
            [8, 12, 6],  # cubic
            [14, 24, 12],  # rhombic dodecahedron xy-square
            [14, 24, 12],  # rhombic dodecahedron xy-hexagon
            [24, 36, 14],  # truncated octahedron
            [12, 18, 8],  # hexagonal cell
        ]
        # iterate through frames and verify # of vertices, edges and faces
        for f in range(5):
            bpy.context.scene.frame_set(f)
            vef = [len(obj_data.vertices), len(obj_data.edges), len(obj_data.polygons)]
            assert vef == expected_vef[f]
        # verify 3x3x3 lattice
        box.show_lattice = True
        for f in range(5):
            bpy.context.scene.frame_set(f)
            vef = [len(obj_data.vertices), len(obj_data.edges), len(obj_data.polygons)]
            np.testing.assert_allclose(vef, np.array(expected_vef[f]) * 27)

    def test_molecule_annotation_molecule_info(self):
        mol = mn.Molecule.load(data_dir / "1cd3.cif")
        assert len(mol.annotations) == 0
        mol.annotations.add_molecule_info()
        assert len(mol.annotations) == 1

    def test_density_annotation_density_info(self, density_file):
        d1 = mn.entities.density.load(density_file)
        assert len(d1.annotations) == 0
        d1.annotations.add_density_info()
        assert len(d1.annotations) == 1

    def test_density_annotation_grid_axes(self, density_file):
        d1 = mn.entities.density.load(density_file)
        assert len(d1.annotations) == 0
        d1.annotations.add_grid_axes()
        assert len(d1.annotations) == 1

    def test_density_annotation_grid_axes_3d(self, density_file):
        d1 = mn.entities.density.load(density_file)
        assert len(d1.annotations) == 0
        d1.annotations.add_grid_axes_3d()
        assert len(d1.annotations) == 1

    def test_common_annotation_label_2d(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        t1.annotations.add_label_2d(text="2D Text", location=(0.25, 0.75))
        assert len(t1.annotations) == 1

    def test_common_annotation_label_3d(self, universe):
        t1 = mn.Trajectory(universe)
        assert len(t1.annotations) == 0
        t1.annotations.add_label_3d(text="3D Text", location=(0.25, 0.5, 0.75))
        assert len(t1.annotations) == 1

    @pytest.mark.skip(reason="This currently fails on MacOS")
    def test_annotations_render_image(self, universe):
        assert "mn_annotations" not in bpy.data.images
        canvas = mn.Canvas(resolution=(192, 108))
        canvas.engine = "CYCLES"  # Only works for this
        canvas.engine.samples = 1
        t1 = mn.Trajectory(mda.Merge(universe.select_atoms("resid 1")))
        t1.annotations.add_com(selection="resid 1")
        bpy.ops.render.render()
        assert mn.scene.compositor.annotations_image in bpy.data.images

    def test_line_mode(self, universe):
        t = mn.Trajectory(universe, name="TestUniverse")
        # test automatic annotation object creation
        name = "MN_an_TestUniverse"
        assert name not in bpy.data.objects
        a = t.annotations.add_com_distance(selection1="resid 1", selection2="resid 2")
        # test no annotation object till geometry gets created
        assert name not in bpy.data.objects
        # enable line mesh to create geometry
        a.line_mode = "mesh"
        assert name in bpy.data.objects
        # test parent
        assert bpy.data.objects[name].parent == t.object
        # verify default material appended
        assert "MN Default" in bpy.data.materials
        # test annotation object creation during update (1)
        bpy.data.objects.remove(bpy.data.objects[name], do_unlink=True)
        assert name not in bpy.data.objects
        t.annotations.visible = True
        assert name in bpy.data.objects
        # test annotation object creation during update (2)
        bpy.data.objects.remove(bpy.data.objects[name], do_unlink=True)
        assert name not in bpy.data.objects
        a.visible = True
        assert name in bpy.data.objects
        # test line mesh enable / disable
        a.line_mode = "overlay"
        # test no vertices when disabled
        assert not bpy.data.objects[name].data.vertices
        # test vertices present when enabled
        a.line_mode = "mesh"
        assert bpy.data.objects[name].data.vertices
        # test attributes (defaults)
        thickness = t.annotations.bob.named_attribute("thickness")
        color = t.annotations.bob.named_attribute("Color")
        # verify single line with arrows at both ends - 5 edges
        assert len(thickness) == 5
        assert len(color) == 5
        np.testing.assert_allclose(thickness, 1.0, rtol=0, atol=0)
        np.testing.assert_allclose(
            color, np.array([(1.0, 1.0, 1.0, 1.0)] * len(color)), rtol=0, atol=0
        )
        # test attribute updates
        a.mesh_thickness = 0.5
        a.mesh_color = (0.0, 1.0, 0.0, 1.0)
        thickness = t.annotations.bob.named_attribute("thickness")
        color = t.annotations.bob.named_attribute("Color")
        np.testing.assert_allclose(thickness, 0.5, rtol=0, atol=0)
        np.testing.assert_allclose(
            color, np.array([(0.0, 1.0, 0.0, 1.0)] * len(color)), rtol=0, atol=0
        )



"""
Auto-generated Style Classes

This file contains Style classes automatically generated from a Blender file.
Each class represents a Style node and provides a Python interface to configure the node parameters.

Generated classes follow the same pattern as the existing styles in molecularnodes.nodes.styles,
using the Socket dataclass system and StyleBase inheritance.
"""

from dataclasses import dataclass
from typing import Any, List, Optional, Tuple
from bpy.types import GeometryNodeGroup

__all__ = [
    "StyleBallAndStick",
    "StyleCartoon",
    "StyleDensitySurface",
    "StyleDensityWire",
    "StylePreset1",
    "StylePreset2",
    "StylePreset3",
    "StylePreset4",
    "StyleRibbon",
    "StyleSpheres",
    "StyleSticks",
    "StyleSurface",
]


@dataclass
class Socket:
    """Represents a mapping between a class attribute and a Blender node input socket.

    Attributes:
        name: The name of the attribute in the Style class
        blendername: The corresponding name of the input socket in the Blender node
    """

    name: str
    blendername: Optional[str] = None


SocketInfo = List[Socket]


class StyleBase:
    style: str
    socketdata: SocketInfo = []

    def update_style_node(self, node_style: GeometryNodeGroup):
        """Update the Blender node inputs with values from this style's attributes.

        Parameters
        ----------
        node_style : GeometryNodeGroup
            The Blender GeometryNodeGroup to update
        """
        for input in node_style.inputs:
            if input.type != "GEOMETRY":
                for arg in self.socketdata:
                    if input.name == arg.blendername:
                        input.default_value = getattr(self, arg.name)


class StyleBallAndStick(StyleBase):
    """Style class for Style Ball and Stick

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    sphere_geometry : Any
        Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
    sphere_radius : float
        Scale the `vdw_radii` attribute before setting the radius for the spheres
    bond_split : bool
        Split apart double and triple bonds visually
    bond_radius : float
        Set the radius for the generated bonds in Angstroms
    bond_find : bool
        Find possible bonds for the selected atoms based on a distance search. Unselected atoms maintain any bonds they already have. Bonds that are found are all treated as single bonds
    bond_find_scale : float
        Scale the VDW radii of the atoms when searching for bonds
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "ball_and_stick"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="sphere_geometry", blendername="Sphere Geometry"),
        Socket(name="sphere_radius", blendername="Sphere Radius"),
        Socket(name="bond_split", blendername="Bond Split"),
        Socket(name="bond_radius", blendername="Bond Radius"),
        Socket(name="bond_find", blendername="Bond Find"),
        Socket(name="bond_find_scale", blendername="Bond Find Scale"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 2,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        sphere_geometry: Any = "Instance",  # Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
        sphere_radius: float = 0.3,  # Scale the `vdw_radii` attribute before setting the radius for the spheres
        bond_split: bool = False,  # Split apart double and triple bonds visually
        bond_radius: float = 0.3,  # Set the radius for the generated bonds in Angstroms
        bond_find: bool = False,  # Find possible bonds for the selected atoms based on a distance search. Unselected atoms maintain any bonds they already have. Bonds that are found are all treated as single bonds
        bond_find_scale: float = 1.0,  # Scale the VDW radii of the atoms when searching for bonds
        color_blur: bool = False,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Ball and Stick

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        sphere_geometry : Any
            Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
        sphere_radius : float
            Scale the `vdw_radii` attribute before setting the radius for the spheres
        bond_split : bool
            Split apart double and triple bonds visually
        bond_radius : float
            Set the radius for the generated bonds in Angstroms
        bond_find : bool
            Find possible bonds for the selected atoms based on a distance search. Unselected atoms maintain any bonds they already have. Bonds that are found are all treated as single bonds
        bond_find_scale : float
            Scale the VDW radii of the atoms when searching for bonds
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.sphere_geometry = sphere_geometry
        self.sphere_radius = sphere_radius
        self.bond_split = bond_split
        self.bond_radius = bond_radius
        self.bond_find = bond_find
        self.bond_find_scale = bond_find_scale
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StyleCartoon(StyleBase):
    """Style class for Style Cartoon

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    peptide_dssp : bool
        Use the DSSP algorithm to compute the `sec_struct` attribute
    peptide_cylinders : bool
        Use cylinders for helices instead of ribbons
    peptide_arrows : bool
        User arrows for sheets
    peptide_rounded : bool
        Create rounded sheets and helices
    peptide_thickness : float
        Thickness for the sheets and helices
    peptide_width : float
        Width for the sheets and helices
    peptide_loop_radius : float
        Radius of the loops for unstructure regions
    peptide_smoothing : float
        Smoothing to apply to sheets
    backbone_shape : Any
        Value for Backbone Shape
    nucleic_width : float
        Value for Nucleic Width
    nucleic_thickness : float
        Value for Nucleic Thickness
    nucleic_radius : float
        Value for Nucleic Radius
    base_shape : Any
        Value for Base Shape
    base_realize : bool
        Value for Base Realize
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "cartoon"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="peptide_dssp", blendername="Peptide DSSP"),
        Socket(name="peptide_cylinders", blendername="Peptide Cylinders"),
        Socket(name="peptide_arrows", blendername="Peptide Arrows"),
        Socket(name="peptide_rounded", blendername="Peptide Rounded"),
        Socket(name="peptide_thickness", blendername="Peptide Thickness"),
        Socket(name="peptide_width", blendername="Peptide Width"),
        Socket(name="peptide_loop_radius", blendername="Peptide Loop Radius"),
        Socket(name="peptide_smoothing", blendername="Peptide Smoothing"),
        Socket(name="backbone_shape", blendername="Backbone Shape"),
        Socket(name="nucleic_width", blendername="Nucleic Width"),
        Socket(name="nucleic_thickness", blendername="Nucleic Thickness"),
        Socket(name="nucleic_radius", blendername="Nucleic Radius"),
        Socket(name="base_shape", blendername="Base Shape"),
        Socket(name="base_realize", blendername="Base Realize"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 2,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        peptide_dssp: bool = False,  # Use the DSSP algorithm to compute the `sec_struct` attribute
        peptide_cylinders: bool = False,  # Use cylinders for helices instead of ribbons
        peptide_arrows: bool = True,  # User arrows for sheets
        peptide_rounded: bool = False,  # Create rounded sheets and helices
        peptide_thickness: float = 0.6,  # Thickness for the sheets and helices
        peptide_width: float = 2.2,  # Width for the sheets and helices
        peptide_loop_radius: float = 0.3,  # Radius of the loops for unstructure regions
        peptide_smoothing: float = 0.5,  # Smoothing to apply to sheets
        backbone_shape: Any = "Cylinder",
        nucleic_width: float = 3.0,
        nucleic_thickness: float = 1.0,
        nucleic_radius: float = 2.0,
        base_shape: Any = "Rectangle",
        base_realize: bool = False,
        color_blur: bool = True,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Cartoon

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        peptide_dssp : bool
            Use the DSSP algorithm to compute the `sec_struct` attribute
        peptide_cylinders : bool
            Use cylinders for helices instead of ribbons
        peptide_arrows : bool
            User arrows for sheets
        peptide_rounded : bool
            Create rounded sheets and helices
        peptide_thickness : float
            Thickness for the sheets and helices
        peptide_width : float
            Width for the sheets and helices
        peptide_loop_radius : float
            Radius of the loops for unstructure regions
        peptide_smoothing : float
            Smoothing to apply to sheets
        backbone_shape : Any
            Value for Backbone Shape
        nucleic_width : float
            Value for Nucleic Width
        nucleic_thickness : float
            Value for Nucleic Thickness
        nucleic_radius : float
            Value for Nucleic Radius
        base_shape : Any
            Value for Base Shape
        base_realize : bool
            Value for Base Realize
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.peptide_dssp = peptide_dssp
        self.peptide_cylinders = peptide_cylinders
        self.peptide_arrows = peptide_arrows
        self.peptide_rounded = peptide_rounded
        self.peptide_thickness = peptide_thickness
        self.peptide_width = peptide_width
        self.peptide_loop_radius = peptide_loop_radius
        self.peptide_smoothing = peptide_smoothing
        self.backbone_shape = backbone_shape
        self.nucleic_width = nucleic_width
        self.nucleic_thickness = nucleic_thickness
        self.nucleic_radius = nucleic_radius
        self.base_shape = base_shape
        self.base_realize = base_realize
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StyleDensitySurface(StyleBase):
    """Style class for Style Density Surface

    Parameters
    ----------
    threshold : float
        Value for Threshold
    shade_smooth : bool
        Apply smooth shading to the created geometry
    hide_dust : float
        Value for Hide Dust
    color : Tuple[float, float, float, float]
        Value for Color
    """

    style = "density_surface"
    socketdata: SocketInfo = [
        Socket(name="threshold", blendername="Threshold"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
        Socket(name="hide_dust", blendername="Hide Dust"),
        Socket(name="color", blendername="Color"),
    ]

    def __init__(
        self,
        threshold: float = 0.8,
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
        hide_dust: float = 0.0,
        color: Tuple[float, float, float, float] = (0.2, 0.51, 0.13, 1.0),
    ):
        """Style class for Style Density Surface

            Parameters
            ----------
        threshold : float
            Value for Threshold
        shade_smooth : bool
            Apply smooth shading to the created geometry
        hide_dust : float
            Value for Hide Dust
        color : Tuple[float, float, float, float]
            Value for Color
        """
        self.threshold = threshold
        self.shade_smooth = shade_smooth
        self.hide_dust = hide_dust
        self.color = color


class StyleDensityWire(StyleBase):
    """Style class for Style Density Wire

    Parameters
    ----------
    threshold : float
        Value for Threshold
    hide_dust : float
        Value for Hide Dust
    wire_radius : float
        Radius of the created wire (in relative nm)
    wire_resolution : int
        Value for Wire Resolution
    color : Tuple[float, float, float, float]
        Value for Color
    """

    style = "density_wire"
    socketdata: SocketInfo = [
        Socket(name="threshold", blendername="Threshold"),
        Socket(name="hide_dust", blendername="Hide Dust"),
        Socket(name="wire_radius", blendername="Wire Radius"),
        Socket(name="wire_resolution", blendername="Wire Resolution"),
        Socket(name="color", blendername="Color"),
    ]

    def __init__(
        self,
        threshold: float = 0.8,
        hide_dust: float = 20.0,
        wire_radius: float = 1.0,  # Radius of the created wire (in relative nm)
        wire_resolution: int = 3,
        color: Tuple[float, float, float, float] = (0.1, 0.39, 0.1, 1.0),
    ):
        """Style class for Style Density Wire

            Parameters
            ----------
        threshold : float
            Value for Threshold
        hide_dust : float
            Value for Hide Dust
        wire_radius : float
            Radius of the created wire (in relative nm)
        wire_resolution : int
            Value for Wire Resolution
        color : Tuple[float, float, float, float]
            Value for Color
        """
        self.threshold = threshold
        self.hide_dust = hide_dust
        self.wire_radius = wire_radius
        self.wire_resolution = wire_resolution
        self.color = color


class StylePreset1(StyleBase):
    """Style class for Style Preset 1

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "preset_1"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur: bool = False,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Preset 1

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StylePreset2(StyleBase):
    """Style class for Style Preset 2

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "preset_2"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur: bool = False,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Preset 2

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StylePreset3(StyleBase):
    """Style class for Style Preset 3

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "preset_3"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Preset 3

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.shade_smooth = shade_smooth


class StylePreset4(StyleBase):
    """Style class for Style Preset 4

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "preset_4"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur: bool = True,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Preset 4

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StyleRibbon(StyleBase):
    """Style class for Style Ribbon

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    backbone_smoothing : float
        Smoothen the sheet ribbons such as beta-sheets
    backbone_threshold : float
        Distance (Angstroms) over which subsequent CA points are treated as a new chain
    backbone_radius : float
        Value for Backbone Radius
    nucleic_backbone_shape : Any
        Value for Nucleic Backbone Shape
    nucleic_backbone_radius : float
        Value for Nucleic Backbone Radius
    backbone_width : float
        Value for Backbone Width
    backbone_thickness : float
        Value for Backbone Thickness
    base_scale : Tuple[float, float, float]
        Value for Base Scale
    base_resolution : int
        Value for Base Resolution
    base_realize : bool
        Value for Base Realize
    uv_map : bool
        Compute and store the `uv_map` for the final protein ribbon geometry
    u_component : Any
        Value for U Component
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "ribbon"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="backbone_smoothing", blendername="Backbone Smoothing"),
        Socket(name="backbone_threshold", blendername="Backbone Threshold"),
        Socket(name="backbone_radius", blendername="Backbone Radius"),
        Socket(name="nucleic_backbone_shape", blendername="Nucleic Backbone Shape"),
        Socket(name="nucleic_backbone_radius", blendername="Nucleic Backbone Radius"),
        Socket(name="backbone_width", blendername="Backbone Width"),
        Socket(name="backbone_thickness", blendername="Backbone Thickness"),
        Socket(name="base_scale", blendername="Base Scale"),
        Socket(name="base_resolution", blendername="Base Resolution"),
        Socket(name="base_realize", blendername="Base Realize"),
        Socket(name="uv_map", blendername="UV Map"),
        Socket(name="u_component", blendername="U Component"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        backbone_smoothing: float = 0.5,  # Smoothen the sheet ribbons such as beta-sheets
        backbone_threshold: float = 4.5,  # Distance (Angstroms) over which subsequent CA points are treated as a new chain
        backbone_radius: float = 1.6,
        nucleic_backbone_shape: Any = "Cylinder",
        nucleic_backbone_radius: float = 1.6,
        backbone_width: float = 3.0,
        backbone_thickness: float = 1.0,
        base_scale: Tuple[float, float, float] = (2.5, 0.5, 7.0),
        base_resolution: int = 4,
        base_realize: bool = False,
        uv_map: bool = False,  # Compute and store the `uv_map` for the final protein ribbon geometry
        u_component: Any = "Factor",
        color_blur: bool = False,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Ribbon

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        backbone_smoothing : float
            Smoothen the sheet ribbons such as beta-sheets
        backbone_threshold : float
            Distance (Angstroms) over which subsequent CA points are treated as a new chain
        backbone_radius : float
            Value for Backbone Radius
        nucleic_backbone_shape : Any
            Value for Nucleic Backbone Shape
        nucleic_backbone_radius : float
            Value for Nucleic Backbone Radius
        backbone_width : float
            Value for Backbone Width
        backbone_thickness : float
            Value for Backbone Thickness
        base_scale : Tuple[float, float, float]
            Value for Base Scale
        base_resolution : int
            Value for Base Resolution
        base_realize : bool
            Value for Base Realize
        uv_map : bool
            Compute and store the `uv_map` for the final protein ribbon geometry
        u_component : Any
            Value for U Component
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.backbone_smoothing = backbone_smoothing
        self.backbone_threshold = backbone_threshold
        self.backbone_radius = backbone_radius
        self.nucleic_backbone_shape = nucleic_backbone_shape
        self.nucleic_backbone_radius = nucleic_backbone_radius
        self.backbone_width = backbone_width
        self.backbone_thickness = backbone_thickness
        self.base_scale = base_scale
        self.base_resolution = base_resolution
        self.base_realize = base_realize
        self.uv_map = uv_map
        self.u_component = u_component
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StyleSpheres(StyleBase):
    """Style class for Style Spheres

    Parameters
    ----------
    geometry : Any
        Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
    radius : float
        Scale the `vdw_radii` of the atom when setting the radius of the spheres
    subdivisions : int
        Number of subdicisions when using _Instances_ or _Mesh_ to represent atoms
    shade_smooth : bool
        Apply smooth shading when using _Instances_ or _Mesh_
    """

    style = "spheres"
    socketdata: SocketInfo = [
        Socket(name="geometry", blendername="Geometry"),
        Socket(name="radius", blendername="Radius"),
        Socket(name="subdivisions", blendername="Subdivisions"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        geometry: Any = "Point",  # Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
        radius: float = 0.8,  # Scale the `vdw_radii` of the atom when setting the radius of the spheres
        subdivisions: int = 2,  # Number of subdicisions when using _Instances_ or _Mesh_ to represent atoms
        shade_smooth: bool = True,  # Apply smooth shading when using _Instances_ or _Mesh_
    ):
        """Style class for Style Spheres

            Parameters
            ----------
        geometry : Any
            Show spheres as a _Point Cloud_, _Instances_ of a mesh Icosphere, or realised _Mesh_ instances of an Icosphere. Point cloud is best for performance and should definitely be used if rendering in Cycles.
        radius : float
            Scale the `vdw_radii` of the atom when setting the radius of the spheres
        subdivisions : int
            Number of subdicisions when using _Instances_ or _Mesh_ to represent atoms
        shade_smooth : bool
            Apply smooth shading when using _Instances_ or _Mesh_
        """
        self.geometry = geometry
        self.radius = radius
        self.subdivisions = subdivisions
        self.shade_smooth = shade_smooth


class StyleSticks(StyleBase):
    """Style class for Style Sticks

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    radius : float
        Radius of the sticks in Angstroms
    color_blur : bool
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "sticks"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="radius", blendername="Radius"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        radius: float = 0.2,  # Radius of the sticks in Angstroms
        color_blur: bool = False,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Sticks

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        radius : float
            Radius of the sticks in Angstroms
        color_blur : bool
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.radius = radius
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth


class StyleSurface(StyleBase):
    """Style class for Style Surface

    Parameters
    ----------
    quality : int
        A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
    scale_radius : float
        Scale the VDW radii of the atoms when creating the surface
    probe_size : float
        Size of the probe that is used to check for solvent accessibility (Angstroms)
    relaxation_steps : int
        Number of times smoothening is applied to the generate surface stretched between the atoms
    separate_by : Any
        Value for Separate By
    group_id : int
        Value for Group ID
    color_source : Any
        Value for Color Source
    color_blur : int
        Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
    shade_smooth : bool
        Apply smooth shading to the created geometry
    """

    style = "surface"
    socketdata: SocketInfo = [
        Socket(name="quality", blendername="Quality"),
        Socket(name="scale_radius", blendername="Scale Radius"),
        Socket(name="probe_size", blendername="Probe Size"),
        Socket(name="relaxation_steps", blendername="Relaxation Steps"),
        Socket(name="separate_by", blendername="Separate By"),
        Socket(name="group_id", blendername="Group ID"),
        Socket(name="color_source", blendername="Color Source"),
        Socket(name="color_blur", blendername="Color Blur"),
        Socket(name="shade_smooth", blendername="Shade Smooth"),
    ]

    def __init__(
        self,
        quality: int = 3,  # A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        scale_radius: float = 1.5,  # Scale the VDW radii of the atoms when creating the surface
        probe_size: float = 1.0,  # Size of the probe that is used to check for solvent accessibility (Angstroms)
        relaxation_steps: int = 10,  # Number of times smoothening is applied to the generate surface stretched between the atoms
        separate_by: Any = "chain_id",
        group_id: int = 0,
        color_source: Any = "Alpha Carbon",
        color_blur: int = 2,  # Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth: bool = True,  # Apply smooth shading to the created geometry
    ):
        """Style class for Style Surface

            Parameters
            ----------
        quality : int
            A lower value results in less geometry, with a higher value meaning better looking but more dense geometry
        scale_radius : float
            Scale the VDW radii of the atoms when creating the surface
        probe_size : float
            Size of the probe that is used to check for solvent accessibility (Angstroms)
        relaxation_steps : int
            Number of times smoothening is applied to the generate surface stretched between the atoms
        separate_by : Any
            Value for Separate By
        group_id : int
            Value for Group ID
        color_source : Any
            Value for Color Source
        color_blur : int
            Interpolate between colors when enabled. When disabled the faces will take their color from their corresponding atom without interpolating
        shade_smooth : bool
            Apply smooth shading to the created geometry
        """
        self.quality = quality
        self.scale_radius = scale_radius
        self.probe_size = probe_size
        self.relaxation_steps = relaxation_steps
        self.separate_by = separate_by
        self.group_id = group_id
        self.color_source = color_source
        self.color_blur = color_blur
        self.shade_smooth = shade_smooth



import os
import typing
from pathlib import Path
from mathutils import Vector
from MDAnalysis.core.groups import AtomGroup
from ...annotations.base import BaseAnnotation
from ...annotations.manager import BaseAnnotationManager
from ..annotations import Label2D, Label3D
from ..base import EntityType

__all__ = [
    "TrajectoryAnnotation",
    "TrajectoryAnnotationManager",
    "AtomInfo",
    "COM",
    "COMDistance",
    "CanonicalDihedrals",
    "UniverseInfo",
]


class TrajectoryAnnotation(BaseAnnotation):
    """
    Base class for a Trajectory Annotation

    All trajectory annotations should derive from this base class and implement
    the 'draw' method. All derived classes will have access to the trajectory
    instance (self.trajectory) and all the annotation inputs and common params
    via self.interface.<property>

    An optional 'defaults' method can be provided to set default values
    to the annotation.

    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Auto-register any sub classes with the annotation manager
        TrajectoryAnnotationManager.register(cls)

    def __init__(self, trajectory):
        # Allow access to the trajectory entity within the annotations
        self.trajectory = trajectory
        super().__init__()


class TrajectoryAnnotationManager(BaseAnnotationManager):
    """
    Annotation Manager for Trajectory Entity

    """

    _entity_type = EntityType.MD
    _classes = {}  # Entity class specific annotation classes

    def __init__(self, entity):
        super().__init__(entity)
        self._interfaces = {}  # Entity instance specific annotation interfaces


class AtomInfo(TrajectoryAnnotation):
    """
    Atom Info Trajectory Annotation

    This annotation shows the atom info of a selection.
    This annotation can be added using the 'add_atom_info' method of the
    trajectory's annotation manager:
        trajectory.annotations.add_atom_info(...)

    Attributes
    ----------
    selection: str
        MDAnalysis selection phrase to select the atom group
    show_resid: bool
        Whether or not to show the res ID along with the atom name
    show_segid: bool
        Whether or not to show the seg ID along with the atom name

    """

    annotation_type = "atom_info"  # required annotation type

    # annotation inputs
    selection: str | AtomGroup = "name CA"
    show_resid: bool = False
    show_segid: bool = False

    def defaults(self) -> None:
        params = self.interface
        # set the default text size and text color
        params.text_size = 16
        params.text_color = (1, 1, 1, 1)

    def validate(self) -> bool:
        params = self.interface
        universe = self.trajectory.universe
        if isinstance(params.selection, str):
            # check if selection phrase is valid
            # mda throws exception if invalid
            self.atom_group = universe.select_atoms(params.selection)
        elif isinstance(params.selection, AtomGroup):
            self.atom_group = params.selection
        else:
            raise ValueError(f"Need str or AtomGroup. Got {type(params.selection)}")
        return True

    def draw(self) -> None:
        params = self.interface
        # iterate over each atom in the atom group
        for atom in self.atom_group:
            text = atom.name
            if params.show_resid:
                text += f"|res {atom.resid}, {atom.resname}"
            if params.show_segid:
                text += f"|seg {atom.segid}"
            self.draw_text_3d(atom.position, text)


class COM(TrajectoryAnnotation):
    """
    Center-of-Mass Trajectory Annotation

    This annotation shows the center-of-mass of a selection.
    This annotation can be added using the 'add_com' method of the
    trajectory's annotation manager:
        trajectory.annotations.add_com(...)

    Attributes
    ----------
    selection: str
        MDAnalysis selection phrase to select the atom group
    text: str
        Text do display at the center-of-mass

    """

    annotation_type = "com"  # required annotation type

    # annotation inputs
    selection: typing.Union[str | AtomGroup] = "protein"
    text: str = "COM"

    def defaults(self) -> None:
        params = self.interface
        # show a default pointer
        params.line_pointer_length = 2

    def validate(self) -> bool:
        params = self.interface
        universe = self.trajectory.universe
        if isinstance(params.selection, str):
            # check if selection phrase is valid
            # mda throws exception if invalid
            self.atom_group = universe.select_atoms(params.selection)
        elif isinstance(params.selection, AtomGroup):
            self.atom_group = params.selection
        else:
            raise ValueError(f"Need str or AtomGroup. Got {type(params.selection)}")
        return True

    def draw(self) -> None:
        com = self.atom_group.center_of_mass()
        self.draw_text_3d(com, self.interface.text)


class COMDistance(TrajectoryAnnotation):
    """
    Distance between Center-of-Masses Trajectory Annotation

    This annotation shows the distance between center-of-masses of two selections.
    This annotation can be added using the 'add_com_distance' method of the
    trajectory's annotation manager:
        trajectory.annotations.add_com_distance(...)

    Attributes
    ----------
    selection1: str
        MDAnalysis selection phrase to select the first atom group
    selection2: str
        MDAnalysis selection phrase to select the second atom group
    text1: str
        Text do display at the first center-of-mass
    text2: str
        Text do display at the second center-of-mass

    """

    annotation_type = "com_distance"  # required annotation type

    # annotation inputs
    selection1: str | AtomGroup
    selection2: str | AtomGroup
    text1: str = "COM1"
    text2: str = "COM2"

    def defaults(self) -> None:
        params = self.interface
        params.line_arrow_size = 0.1

    def validate(self) -> bool:
        params = self.interface
        universe = self.trajectory.universe
        # check selection 1
        if isinstance(params.selection1, str):
            # check if selection phrase is valid
            # mda throws exception if invalid
            self.atom_group1 = universe.select_atoms(params.selection1)
        elif isinstance(params.selection1, AtomGroup):
            self.atom_group1 = params.selection1
        else:
            raise ValueError(f"Need str or AtomGroup. Got {type(params.selection1)}")
        # check selection 2
        if isinstance(params.selection2, str):
            # check if selection phrase is valid
            # mda throws exception if invalid
            self.atom_group2 = universe.select_atoms(params.selection2)
        elif isinstance(params.selection2, AtomGroup):
            self.atom_group2 = params.selection2
        else:
            raise ValueError(f"Need str or AtomGroup. Got {type(params.selection2)}")
        return True

    def draw(self) -> None:
        params = self.interface
        # calculate the two center-of-masses
        com1 = self.atom_group1.center_of_mass()
        com2 = self.atom_group2.center_of_mass()
        # calculate distance between coms
        distance = self.distance(com1, com2)
        # draw line with arrow ends showing distance in between
        self.draw_line_3d(
            v1=com1,
            v2=com2,
            v1_text=params.text1,
            v2_text=params.text2,
            mid_text=f"{distance:1.2f} Å",
            v1_arrow=True,
            v2_arrow=True,
        )


class CanonicalDihedrals(TrajectoryAnnotation):
    """
    Canonical Dihedrals of a given residue

    Attributes
    ----------
    resid: int
        The residue number

    show_atom_names: bool
        Whether or not to show the individual atom names in the residue

    show_direction: bool
        Whether or not to show the arc indicating the angle direction

    """

    annotation_type = "canonical_dihedrals"

    resid: int
    show_atom_names: bool = True
    show_direction: bool = True

    def defaults(self) -> None:
        params = self.interface
        params.line_arrow_size = 10.0
        params.mesh_thickness = 0.1

    def validate(self) -> bool:
        params = self.interface
        universe = self.trajectory.universe
        # verify residue in universe - raises exception if invalid
        self.residue = universe.residues[params.resid - 1]
        return True

    def _get_start_dv(self, dihedral: list) -> Vector:
        """Get the direction vector along which to start the arc"""
        v0 = Vector(dihedral.atoms[0].position)
        v1 = Vector(dihedral.atoms[1].position)
        v2 = Vector(dihedral.atoms[2].position)
        # find two vectors on the first plane (between 0, 1, 2)
        v10 = v0 - v1
        v20 = v0 - v2
        # the normal of the plane on which the arc is drawn
        normal = v2 - v1
        # (v10 x normal) x v20 is a vector in the first plane that is
        # perpendicular to the normal - get the unit vector
        dv = v10.cross(normal).cross(v20)
        dv.normalize()
        return dv

    def draw(self) -> None:
        params = self.interface
        r = self.residue
        # *_selection() returns None if nothing found
        selections = [
            r.phi_selection(),
            r.psi_selection(),
            r.omega_selection(),
            r.chi1_selection(),
        ]
        symbols = ["ϕ", "ψ", "ω", "χ1"]
        # iterate over the canonical selections
        for si in range(4):
            if selections[si] is None:
                continue
            # iterate over the four atoms in the selection
            for ai in range(4):
                a1 = selections[si].atoms[ai]
                if params.show_atom_names:
                    self.draw_text_3d(a1.position, a1.name)
                if ai == 3:
                    continue
                a2 = selections[si].atoms[ai + 1]
                v1 = a1.position
                v2 = a2.position
                text = None
                if ai == 1:
                    angle = selections[si].dihedral.value()
                    text = f"{symbols[si]} = {angle:1.2f} °"
                    if params.show_direction:
                        center = (v1 + v2) / 2
                        radius = 0.1 * self.distance(v1, v2)  # 10% of distance
                        # draw arc indicating direction of the angle
                        start_dv = self._get_start_dv(selections[si])
                        self.draw_circle_3d(
                            center,
                            radius,
                            (v2 - v1),
                            angle=angle,
                            start_dv=start_dv,
                            c_arrow=True,
                        )
                self.draw_line_3d(v1, v2, mid_text=text)


class UniverseInfo(TrajectoryAnnotation):
    """
    Universe Info Trajectory Annotation

    Attributes
    ----------
    location: tuple[float, float]
        Normalized coordinates (0.0 - 1.0) of the postion in viewport / render

    show_frame: bool
        Whether or not to show the frame number

    show_topology: bool
        Whether or not to show the topology filename

    show_trajectory: bool
        Whether or not to show the trajectory filename

    show_atoms: bool
        Whether or not to show the number of atoms

    custom_text: str
        Any custom text to add at the end of the annotation

    """

    annotation_type = "universe_info"

    location: tuple[float, float] = (0.025, 0.05)
    show_frame: bool = True
    show_topology: bool = True
    show_trajectory: bool = True
    show_atoms: bool = True
    custom_text: str = ""

    def defaults(self) -> None:
        params = self.interface
        params.text_align = "left"

    def validate(self) -> bool:
        params = self.interface
        x, y = params.location
        if (not 0 <= x <= 1) or (not 0 <= y <= 1):
            raise ValueError("Normalized coordinates should lie between 0 and 1")
        return True

    def draw(self) -> None:
        params = self.interface
        u = self.trajectory.universe
        text = ""
        if params.show_frame:
            text = f"Frame : {u.trajectory.frame} / {u.trajectory.n_frames - 1}"
        if params.show_topology and isinstance(u.filename, (str, Path)):
            text = text + "|Topology : " + os.path.basename(u.filename)
        if params.show_trajectory and isinstance(u.trajectory.filename, (str, Path)):
            text = text + "|Trajectory : " + os.path.basename(u.trajectory.filename)
        if params.show_atoms:
            text = text + "|Atoms : " + str(u.trajectory.n_atoms)
        if params.custom_text != "":
            text = text + "|" + params.custom_text
        # Draw text at normalized coordinates wrt viewport / render
        self.draw_text_2d_norm(params.location, text)


class SimulationBox(TrajectoryAnnotation):
    """
    Simulation Box of Trajectory (if present)

    Attributes
    ----------
    center_to_origin: bool
        Whether to move the center of the box to Origin (0, 0, 0)

    compact: bool
        Whether to create a compact Wigner-Seitz cell
        Note that the trajectory needs to have been wrapped with the
        compact option to see it within the box

    show_lattice: bool
        Whether to show a 3x3x3 lattice of the box

    """

    annotation_type = "simulation_box"

    center_to_origin: bool = False
    compact: bool = False
    show_lattice: bool = False

    def defaults(self) -> None:
        params = self.interface
        params.mesh_wireframe = True
        params.mesh_thickness = 5.0
        params.mesh_shade_smooth = False

    def draw(self) -> None:
        params = self.interface
        u = self.trajectory.universe
        ts = u.trajectory.ts
        # not all universes have dimensions set
        if ts.dimensions is not None:
            # draw box
            if params.compact:
                # Compact Wigner-Seitz cell
                self.draw_wigner_seitz_cell(
                    ts.triclinic_dimensions,
                    params.center_to_origin,
                    params.show_lattice,
                )
            else:
                # Regular Triclinic cell
                a, b, c, alpha, beta, gamma = ts.dimensions
                origin = Vector((0, 0, 0))
                # move box center to origin if set
                if params.center_to_origin:
                    origin = Vector(-sum(ts.triclinic_dimensions) / 2)
                self.draw_triclinic_cell(
                    a,
                    b,
                    c,
                    alpha,
                    beta,
                    gamma,
                    origin=origin,
                    show_lattice=params.show_lattice,
                )


class Label2D(TrajectoryAnnotation, Label2D):
    """Common Label2D Annotation for all entities"""


class Label3D(TrajectoryAnnotation, Label3D):
    """Common Label3D Annotation for all entities"""



import bpy
import numpy as np
from bpy.app.handlers import persistent
from PIL import Image
from .annotations.utils import render_annotations
from .scene.compositor import annotations_image


# this update function requires a self and context input, as funcitons with these inputs
# have ot be passed to the `update` arguments of UI properties. When the UI is updated,
# the function is called with the UI element being `self` and the current context being
# passed into the function
def _update_entities(self, context: bpy.types.Context) -> None:
    """
    Function for being called at various points in the updating of the UI, to ensure
    positions and selections of the trajectories are udpated with the new inputs
    """
    update_entities(context.scene)


def _selection_update_trajectories(self, context: bpy.types.Context) -> None:
    """
    Function for selection changing. If the selection is immutable e.g.
    when it is generated from an AtomGroup,
    update universe will not be called to avoid invalid selection str.
    """
    if self.immutable:
        return
    else:
        update_entities(context.scene)


# this is the 'perisisent' function which can be appended onto the
# `bpy.app.handlers.frame_change_*` functions. Either before or after the frame changes
# this function will then be called - ensuring all of the trajectories are up to date. We
# use the `frame_change_pre` handler as we want the frame to change first, then we update
# the universe based on the current frame value
@persistent
def update_entities(scene):
    "Call the `set_frame()` method of all entities in the current session"
    session = scene.MNSession
    session.prune()

    for entity in session.entities.values():
        if entity.update_with_scene:
            frame_to_set = scene.frame_current
        else:
            frame_to_set = entity.frame

        # do the entity setting, if the method isn't implemented, just pass
        try:
            entity.set_frame(frame_to_set)
        except NotImplementedError:
            pass


@persistent
def render_pre_handler(scene: bpy.types.Scene) -> None:
    """
    Blender's on render (before) handler
    Any changes needed before the rendering of a frame need to go in here

    """
    # frame_change_pre is supposed to be called before render_pre
    # however, that doesn't seem to be the case
    # update entities here to ensure annotations match entity frame
    update_entities(scene)
    # Render annotations to an image
    bpy.context.view_layer.update()
    render_scale = scene.render.resolution_percentage / 100
    width = int(scene.render.resolution_x * render_scale)
    height = int(scene.render.resolution_y * render_scale)
    image_scale = 2  # To workaround anti-aliasing issues with lines
    image = Image.new(
        "RGBA",
        (width * image_scale, height * image_scale),
        (0, 0, 0, 0),
    )
    # render annotations of all entities
    render_annotations(scene, render_scale, image, image_scale)
    if image_scale != 1:
        # scale down to actual render size
        image = image.resize((width, height))
    # create blender image
    if annotations_image not in bpy.data.images:
        bpy.data.images.new(annotations_image, width, height)
    bpy_image = bpy.data.images[annotations_image]
    bpy_image.scale(width, height)
    # update from PIL image
    bpy_image.pixels[:] = (np.flipud(np.array(image).astype(float)) / 255.0).ravel()



import os
import subprocess
import sys

argv = sys.argv
argv = argv[argv.index("--") + 1 :]


def main():
    python = os.path.realpath(sys.executable)
    subprocess.run([python] + argv)


if __name__ == "__main__":
    main()



"""
Python module for converting molecular structures from biotite into Blender objects.

This module provides functions to:
1. Create 3D models from biotite AtomArray objects (`create_object`)
2. Attach molecular information as attributes to Blender objects (`atom_array_to_named_attributes`)

The created Blender objects can be used for visualization and manipulation of molecular
structures within Blender's 3D environment. Properties like atom positions, bonds, and
other molecular attributes are preserved as named attributes that can be used with
Geometry Nodes and other Blender features.
"""

from typing import Any, Callable
import bpy
import databpy as db
import numpy as np
from biotite.structure import AtomArray, AtomArrayStack

__all__ = ["create_object", "atom_array_to_named_attributes"]


def _get_gn_modifier(obj: bpy.types.Object, name: str) -> bpy.types.NodesModifier:
    mod = obj.modifiers[name]
    if not isinstance(mod, bpy.types.NodesModifier):
        raise TypeError(
            f"Modifier {name} is of type {type(mod)} not {bpy.types.NodesModifier}"
        )
    return mod


def _unique_aname(obj: bpy.types.Object, prefix: str = "sel") -> str:
    attributes = db.list_attributes(obj)
    counter = 0
    aname = "{}_{}".format(prefix, counter)
    while aname in attributes:
        counter += 1
        aname = "{}_{}".format(prefix, counter)

    return aname


def _validate_non_negative(value: int) -> None:
    """Validate non-negative integers.

    Parameters
    ----------
    value : int
        Value to validate

    Raises
    ------
    ValueError
        If value is negative
    """
    if value < 0:
        raise ValueError(f"Value must be non-negative, got {value}")


class ObjectMNProperty:
    """Descriptor for Blender object properties with optional validation.

    Provides clean interface for Blender custom properties with validation support.

    Examples
    --------
    >>> class MyClass:
    ...     frame = ObjectMNProperty("frame", validate_fn=lambda x: x >= 0)
    """

    def __init__(
        self,
        attr_name: str,
        validate_fn: Callable[[Any], None] | None = None,
    ):
        """Initialize the descriptor.

        Parameters
        ----------
        attr_name : str
            Name of the property on the Blender object
        validate_fn : callable, optional
            Validation function that raises on invalid values
        """
        self.attr_name = attr_name
        self.validate_fn = validate_fn

    def __set_name__(self, owner, name):
        """Store the Python attribute name."""
        self.name = name

    def __get__(self, obj, objtype=None):
        """Get property value from Blender object."""
        return getattr(obj.object.mn, self.attr_name)

    def __set__(self, obj, value):
        """Set property value on Blender object with validation."""
        if self.validate_fn is not None:
            self.validate_fn(value)
        setattr(obj.object.mn, self.attr_name, value)


class IntObjectMNProperty(ObjectMNProperty):
    def __get__(self, obj, objtype=None) -> int:
        return super().__get__(obj, objtype)

    def __set__(self, obj, value: int):
        return super().__set__(obj, value)


class BoolObjectMNProperty(ObjectMNProperty):
    def __get__(self, obj, objtype=None) -> bool:
        return super().__get__(obj, objtype)

    def __set__(self, obj, value: bool):
        return super().__set__(obj, value)


class StringObjectMNProperty(ObjectMNProperty):
    def __get__(self, obj, objtype=None) -> str:
        return super().__get__(obj, objtype)

    def __set__(self, obj, value: str):
        return super().__set__(obj, value)


def create_object(
    array: AtomArray | AtomArrayStack,
    name="NewObject",
    centre: str | None = None,
    world_scale: float = 0.01,
    collection: bpy.types.Collection | None = None,
) -> bpy.types.Object:
    """
    Create a 3D model of the molecule, with one vertex for each atom.

    This function converts a biotite AtomArray into a Blender object with vertices
    representing atoms and edges representing bonds. All atomic properties from the
    AtomArray are stored as named attributes on the Blender mesh, making them
    accessible for use with Geometry Nodes and other Blender features.

    Parameters
    ----------
    array : AtomArray
        The molecular structure to convert into a Blender object.
    name : str, optional
        Name of the created Blender object. Default is "NewObject".
    centre : str or None, optional
        If provided, centers the object according to the specified criteria.
        Default is None (no centering).
    world_scale : float, optional
        Scale factor to apply to atomic coordinates when creating the model.
        Default is 0.01, which converts Ångströms to a reasonable size in Blender.
    collection : bpy.types.Collection or None, optional
        The Blender collection to add the new object to. If None, the object
        is added to the active collection. Default is None.

    Returns
    -------
    bpy.types.Object
        The created 3D model, as an object in the 3D scene.
        The object contains vertices for atoms and edges for bonds,
        with all molecular properties stored as named attributes.
    """

    if isinstance(array, AtomArrayStack):
        array = array[0]

    bob = db.create_bob(
        vertices=array.coord * world_scale,
        edges=array.bonds.as_array()[:, :2] if array.bonds is not None else None,
        name=name,
        collection=collection,
    )
    # Add information about the bond types to the model on the edge domain
    # Bond types: 'ANY' = 0, 'SINGLE' = 1, 'DOUBLE' = 2, 'TRIPLE' = 3, 'QUADRUPLE' = 4
    # 'AROMATIC_SINGLE' = 5, 'AROMATIC_DOUBLE' = 6, 'AROMATIC_TRIPLE' = 7
    # https://www.biotite-python.org/apidoc/biotite.structure.BondType.html#biotite.structure.BondType
    if array.bonds:
        bob.store_named_attribute(
            array.bonds.as_array()[:, 2],
            "bond_type",
            domain=db.AttributeDomains.EDGE,
            atype=db.AttributeTypes.INT,
        )

    atom_array_to_named_attributes(array, bob.object, world_scale=world_scale)

    if centre is not None:
        bob.position -= bob.centroid(centre)

    return bob.object


def atom_array_to_named_attributes(
    array: AtomArray, obj: bpy.types.Object, world_scale: float = 0.01
) -> None:
    """
    Store all annotations from an AtomArray as named attributes on Blender vertex data.

    This function iterates through the annotations (properties) of a biotite AtomArray
    and stores them as named attributes on a Blender mesh object. These attributes can
    then be accessed and manipulated using Blender's Geometry Nodes system or through Python.

    Only numeric and boolean attributes are stored, as Blender's Geometry Nodes doesn't
    currently support string attributes. String attributes from the AtomArray should have
    been converted to numeric representations during the import process.

    Parameters
    ----------
    array : AtomArray
        The biotite AtomArray containing molecular data and annotations to be stored.
    obj : bpy.types.Object
        The Blender object (typically a mesh) on which to store the attributes.
    world_scale : float, optional
        Scale factor to apply to size-related values like van der Waals radii.
        Default is 0.01 to convert from Ångströms to Blender units.

    Notes
    -----
    - Coordinate data ('coord') is skipped as it's already stored as vertex positions.
    - 'hetero' flag is also skipped from the annotations.
    - Any string attributes that were converted to integer codes will have the '_int'
      suffix removed when stored on the mesh.
    """

    # don't need to add coordinates as those have been stored as `position` on the mesh
    annotations_to_skip = ["coord", "hetero"]

    for attr in array.get_annotation_categories():
        if attr in annotations_to_skip:
            continue

        data = array.get_annotation(attr)  # type: ignore

        if attr == "vdw_radii":
            data *= world_scale

        # geometry nodes doesn't support strings at the moment so we can only store the
        # numeric and boolean attributes on the mesh. All string attributes should have
        # already been converted to a corresponding numeric attribute during the
        # reader process

        if not (
            np.issubdtype(data.dtype, np.number) or np.issubdtype(data.dtype, np.bool_)
        ):
            continue

        # the integer versions of strings have been added as annotations that just
        # append `_int` onto the name so we don't overrite the original data but when
        # storing on the meshh we can just remove the `_int`
        db.store_named_attribute(
            obj=obj,
            data=data,
            name=attr.replace("_int", ""),
        )



from abc import ABCMeta
from pathlib import Path
from typing import Union
import bpy
from ... import blender as bl
from ..base import EntityType, MolecularEntity


class Ensemble(MolecularEntity, metaclass=ABCMeta):
    def __init__(self, file_path: Union[str, Path]):
        """
        Initialize an Ensemble object.

        Parameters
        ----------
        file_path : Union[str, Path]
            The path to the file.
        """
        super().__init__()
        self._entity_type = EntityType.ENSEMBLE
        self.type: str = "ensemble"
        self.file_path: Path = bl.path_resolve(file_path)

    @property
    def instance_collection(self) -> bpy.types.Collection:
        """
        The instances of the ensemble.

        Returns
        -------
        bpy.types.Collection
            The collection containing the ensemble instances.
        """
        return bpy.data.collections[self._instance_collection_name]

    @instance_collection.setter
    def instance_collection(self, value: bpy.types.Collection) -> None:
        """
        Set the instance collection.

        Parameters
        ----------
        value : bpy.types.Collection
            The collection to set as the instance collection.

        Raises
        ------
        ValueError
            If the value is not a bpy.types.Collection.
        """
        if not isinstance(value, bpy.types.Collection):
            raise ValueError("The instances must be a bpy.types.Collection.")
        self._instance_collection_name = value.name

    def create_object(
        self,
        name: str = "NewEnsemble",
        node_setup: bool = True,
        world_scale: float = 0.01,
        fraction: float = 1.0,
        simplify=False,
    ) -> bpy.types.Object:
        """
        Create a 3D object for the ensemble.

        Parameters
        ----------
        name : str, optional
            The name of the model, by default "NewEnsemble"
        node_setup : bool, optional
            Whether to setup nodes for the data and instancing objects, by default True
        world_scale : float, optional
            Scaling transform for the coordinates before loading in to Blender, by default 0.01
        fraction : float, optional
            The fraction of the instances to display on loading. Reducing can help with performance, by default 1.0
        simplify : bool, optional
            Whether to instance the given models or simplify them for debugging and performance, by default False

        Notes
        -----
        Creates a data object which stores all of the required instancing information. If
        there are molecules to be instanced, they are also created in their own data collection.
        """
        pass



import molecularnodes as mn


def test_select_ca():
    mol = mn.Molecule.fetch("8H1B")
    mol.select.atom_name("CA").chain_id(["A"]).store_selection("custom_selection")
    assert sum(mol.named_attribute("custom_selection")) == 190
    mol.select.reset().atom_name("CA").store_selection("custom_selection")
    assert sum(mol.named_attribute("custom_selection")) == 384
    (
        mol.select.reset()
        .is_peptide_backbone()
        .res_id(range(50))
        .chain_id(["B"])
        .res_name(["LYS", "ARG", "HIS"])
        .store_selection("custom_selection")
    )
    assert sum(mol.named_attribute("custom_selection")) == 18


def test_evalute_on_array():
    mol = mn.Molecule.fetch("8H1B")
    sel = mn.entities.MoleculeSelector()
    sel.atom_name("CA").chain_id(["A"]).res_id(range(50))

    assert sum(sel.evaluate_on_array(mol.array)) == 49


def test_adjust_seletion():
    mol = mn.Molecule.fetch("8H1B")
    sel = mol.select

    (
        sel.res_id(range(200))
        .chain_id(["A", "B"])
        .res_name(["LYS", "ARG", "HIS", "GLY", "PRO"])
        .is_backbone()
    )

    assert sum(sel.evaluate_on_array(mol.array)) == 400
    sel.res_id(range(100, 200))
    assert sum(sel.evaluate_on_array(mol.array)) == 192
